from typing import List, Optional

from pygitguardian.iac_models import IaCFileResult, IaCScanResult

from ggshield.verticals.iac.collection.filter_ignored import filter_unignored_files
from ggshield.verticals.iac.collection.iac_scan_collection import (
    CollectionType,
    IaCScanCollection,
)


class IaCPathScanCollection(IaCScanCollection[IaCScanResult]):
    type = CollectionType.PathScan

    @property
    def has_results(self) -> bool:
        return self.result is not None and bool(self.result.entities_with_incidents)

    def get_entities_without_ignored(self) -> Optional[List[IaCFileResult]]:
        if self.result is None:
            return None

        return filter_unignored_files(self.result.entities_with_incidents)

    def get_result_without_ignored(self) -> Optional[IaCScanResult]:
        entities_without_ignored = self.get_entities_without_ignored()
        if self.result is None or entities_without_ignored is None:
            return None

        result_dict = self.result.to_dict()
        del result_dict["entities_with_incidents"]
        return IaCScanResult(
            **result_dict, entities_with_incidents=entities_without_ignored
        )



from pathlib import Path
from typing import Any, Optional, Sequence

import click

from ggshield.cmd.sca.scan.sca_scan_utils import (
    create_output_handler,
    sca_scan_all,
    sca_scan_diff,
)
from ggshield.cmd.sca.scan.scan_common_options import (
    add_sca_scan_common_options,
    update_context,
)
from ggshield.cmd.utils.common_decorators import exception_wrapper
from ggshield.cmd.utils.common_options import all_option, directory_argument
from ggshield.cmd.utils.hooks import check_user_requested_skip
from ggshield.core.scan.scan_mode import ScanMode
from ggshield.verticals.sca.collection.collection import (
    SCAScanAllVulnerabilityCollection,
    SCAScanDiffVulnerabilityCollection,
)


@click.command()
@add_sca_scan_common_options()
@all_option
@directory_argument
@click.pass_context
@exception_wrapper
def scan_pre_commit_cmd(
    ctx: click.Context,
    exit_zero: bool,
    minimum_severity: str,
    ignore_paths: Sequence[str],
    ignore_fixable: bool,
    ignore_not_fixable: bool,
    scan_all: bool,
    directory: Optional[Path],
    **kwargs: Any,
) -> int:
    """
    Scans if the currently staged files introduce SCA vulnerabilities.

    This command checks if the currently staged files introduce SCA vulnerabilities
    compared to the current state of the repository.

    Scanning a repository with this command will not trigger any incident on your dashboard.

    Only metadata such as call time, request size and scan mode is stored server-side.
    """
    if check_user_requested_skip():
        return 0

    if directory is None:
        directory = Path().resolve()

    # Adds client and required parameters to the context
    update_context(
        ctx,
        exit_zero,
        minimum_severity,
        ignore_paths,
        ignore_fixable,
        ignore_not_fixable,
    )

    if scan_all:
        result = sca_scan_all(ctx, directory, scan_mode=ScanMode.PRE_COMMIT_ALL)
        scan = SCAScanAllVulnerabilityCollection(id=str(directory), result=result)
        output_handler = create_output_handler(ctx)
        return output_handler.process_scan_all_result(scan)

    result = sca_scan_diff(
        ctx=ctx,
        directory=directory,
        previous_ref="HEAD",
        include_staged=True,
        scan_mode=ScanMode.PRE_COMMIT_DIFF,
    )

    scan = SCAScanDiffVulnerabilityCollection(id=str(directory), result=result)
    output_handler = create_output_handler(ctx)
    return output_handler.process_scan_diff_result(scan)



"""
This module defines options which should be available on all iac scan subcommands.

To use it:
- Add the `@add_common_iac_scan_options()` decorator after all the `click.option()`
    calls of the command function.
- Add a `**kwargs: Any` argument to the command function.

The `kwargs` argument is required due to the way click works,
`add_common_options()` adds an argument for each option it defines.
"""

from typing import Any, Callable, Sequence

import click

from ggshield.cmd.utils.common_options import (
    AnyFunction,
    add_common_options,
    exit_zero_option,
    ignore_path_option,
    json_option,
    minimum_severity_option,
)
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.core.client import create_client_from_config
from ggshield.core.config.user_config import (
    POLICY_ID_PATTERN,
    IaCConfigIgnoredPath,
    IaCConfigIgnoredPolicy,
    validate_policy_id,
)
from ggshield.core.filter import init_exclusion_regexes


def _validate_exclude(_ctx: Any, _param: Any, value: Sequence[str]) -> Sequence[str]:
    invalid_excluded_policies = [
        policy_id for policy_id in value if not validate_policy_id(policy_id)
    ]
    if len(invalid_excluded_policies) > 0:
        raise ValueError(
            f"The policies {invalid_excluded_policies} do not match the pattern '{POLICY_ID_PATTERN.pattern}'"
        )
    return value


_ignore_policy_option = click.option(
    "--ignore-policy",
    "--ipo",
    "ignore_policies",
    multiple=True,
    help="Policies to exclude from the results.",
    callback=_validate_exclude,
)


def add_iac_scan_common_options() -> Callable[[AnyFunction], AnyFunction]:
    def decorator(cmd: AnyFunction) -> AnyFunction:
        add_common_options()(cmd)
        exit_zero_option(cmd)
        minimum_severity_option(cmd)
        _ignore_policy_option(cmd)
        ignore_path_option(cmd)
        json_option(cmd)
        return cmd

    return decorator


def update_context(
    ctx: click.Context,
    exit_zero: bool,
    minimum_severity: str,
    ignore_policies: Sequence[str],
    ignore_paths: Sequence[str],
) -> None:
    ctx_obj = ContextObj.get(ctx)
    config = ctx_obj.config
    ctx_obj.client = create_client_from_config(config, ctx_obj.ui)

    if ignore_paths is not None:
        config.user_config.iac.ignored_paths.extend(
            (IaCConfigIgnoredPath(path=path) for path in ignore_paths)
        )

    ctx_obj.exclusion_regexes = init_exclusion_regexes(
        {ignored.path for ignored in config.user_config.iac.ignored_paths}
    )

    if ignore_policies is not None:
        config.user_config.iac.ignored_policies.extend(
            (IaCConfigIgnoredPolicy(policy=policy) for policy in ignore_policies)
        )

    if exit_zero is not None:
        config.user_config.exit_zero = exit_zero

    if minimum_severity is not None:
        config.user_config.iac.minimum_severity = minimum_severity



from pathlib import Path
from typing import Any, Optional, Sequence

import click

from ggshield.cmd.iac.scan.all import display_iac_scan_all_result, iac_scan_all
from ggshield.cmd.iac.scan.diff import display_iac_scan_diff_result, iac_scan_diff
from ggshield.cmd.iac.scan.iac_scan_common_options import (
    add_iac_scan_common_options,
    update_context,
)
from ggshield.cmd.iac.scan.iac_scan_utils import augment_unignored_issues
from ggshield.cmd.utils.common_decorators import display_beta_warning, exception_wrapper
from ggshield.cmd.utils.common_options import directory_argument
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.core.git_hooks.ci.get_scan_ci_parameters import (
    NotAMergeRequestError,
    get_scan_ci_parameters,
)
from ggshield.core.git_hooks.ci.supported_ci import SupportedCI
from ggshield.core.scan.scan_mode import ScanMode
from ggshield.utils.git_shell import git


@click.command()
@add_iac_scan_common_options()
@directory_argument
@click.pass_context
@display_beta_warning
@exception_wrapper
def scan_ci_cmd(
    ctx: click.Context,
    exit_zero: bool,
    minimum_severity: str,
    ignore_policies: Sequence[str],
    ignore_paths: Sequence[str],
    directory: Optional[Path] = None,
    **kwargs: Any,
) -> int:
    """
    Scan in CI for IaC vulnerabilities. By default, it will return vulnerabilities added in the new commits.

    The scan is successful if no *new* IaC vulnerability was found, unless `--all` is used,
    in which case the scan is only successful if no IaC vulnerability (old and new) was found.
    """
    if directory is None:
        directory = Path().resolve()

    update_context(ctx, exit_zero, minimum_severity, ignore_policies, ignore_paths)
    config = ContextObj.get(ctx).config
    ci_mode = SupportedCI.from_ci_env()

    try:
        # we will work with branch names and deep commits, so we run a git fetch to ensure the
        # branch names and commit sha are locally available
        git(["fetch"], cwd=directory)
        params = get_scan_ci_parameters(
            ci_mode, wd=directory, verbose=config.user_config.verbose
        )
        if params is None:
            click.echo(
                "No commit found in merge request, skipping scan.",
                err=True,
            )
            return 0

        current_commit, reference_commit = params

        result = iac_scan_diff(
            ctx,
            directory,
            reference_commit,
            current_ref=current_commit,
            include_staged=True,
            scan_mode=ScanMode.CI_DIFF,
            ci_mode=ci_mode,
        )
        augment_unignored_issues(config.user_config, result)
        return display_iac_scan_diff_result(ctx, directory, result)
    except NotAMergeRequestError:
        click.echo(
            (
                "WARNING: scan ci expects to be run in a merge-request pipeline.\n"
                "No target branch could be identified, will perform a scan all instead.\n"
                "This is a fallback behaviour, that will be removed in a future version."
            ),
            err=True,
        )
        result = iac_scan_all(
            ctx, directory, scan_mode=ScanMode.CI_ALL, ci_mode=ci_mode
        )
        augment_unignored_issues(config.user_config, result)
        return display_iac_scan_all_result(ctx, directory, result)



from typing import Any

import click

from ggshield.cmd.utils.common_options import add_common_options

from .config_get import config_get_cmd
from .config_list import config_list_cmd
from .config_migrate import config_migrate_cmd
from .config_set import config_set_cmd
from .config_unset import config_unset_cmd


@click.group(
    commands={
        "list": config_list_cmd,
        "set": config_set_cmd,
        "unset": config_unset_cmd,
        "get": config_get_cmd,
        "migrate": config_migrate_cmd,
    }
)
@add_common_options()
def config_group(**kwargs: Any) -> None:
    """Commands to manage configuration."""



from pathlib import Path
from typing import Any, Optional, Sequence

import click

from ggshield.cmd.sca.scan.sca_scan_utils import (
    create_output_handler,
    sca_scan_all,
    sca_scan_diff,
)
from ggshield.cmd.sca.scan.scan_common_options import (
    add_sca_scan_common_options,
    update_context,
)
from ggshield.cmd.utils.common_decorators import exception_wrapper
from ggshield.cmd.utils.common_options import directory_argument
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.core.git_hooks.ci.get_scan_ci_parameters import (
    NotAMergeRequestError,
    get_scan_ci_parameters,
)
from ggshield.core.git_hooks.ci.supported_ci import SupportedCI
from ggshield.core.scan.scan_mode import ScanMode
from ggshield.utils.git_shell import git
from ggshield.verticals.sca.collection.collection import (
    SCAScanAllVulnerabilityCollection,
    SCAScanDiffVulnerabilityCollection,
)


@click.command()
@add_sca_scan_common_options()
@click.pass_context
@directory_argument
@exception_wrapper
def scan_ci_cmd(
    ctx: click.Context,
    exit_zero: bool,
    minimum_severity: str,
    ignore_paths: Sequence[str],
    ignore_fixable: bool,
    ignore_not_fixable: bool,
    directory: Optional[Path],
    **kwargs: Any,
) -> int:
    """
    Evaluates if a CI event introduces SCA vulnerabilities.

    Scanning a repository with this command will not trigger any incident on your dashboard.

    Only metadata such as call time, request size and scan mode is stored server-side.
    """
    if directory is None:
        directory = Path().resolve()

    # Adds client and required parameters to the context
    update_context(
        ctx,
        exit_zero,
        minimum_severity,
        ignore_paths,
        ignore_fixable,
        ignore_not_fixable,
    )

    config = ContextObj.get(ctx).config
    ci_mode = SupportedCI.from_ci_env()
    output_handler = create_output_handler(ctx)

    try:
        # we will work with branch names and deep commits, so we run a git fetch to ensure the
        # branch names and commit sha are locally available
        git(["fetch"], cwd=directory)
        params = get_scan_ci_parameters(
            ci_mode, wd=directory, verbose=config.user_config.verbose
        )
        if params is None:
            click.echo(
                "No commit found in merge request, skipping scan.",
                err=True,
            )
            return 0

        current_commit, reference_commit = params
        scan_mode = f"{ScanMode.CI_DIFF.value}/{ci_mode.value}"
        result = sca_scan_diff(
            ctx=ctx,
            directory=directory,
            previous_ref=reference_commit,
            current_ref=current_commit,
            scan_mode=scan_mode,
            ci_mode=ci_mode.name,
        )
        scan = SCAScanDiffVulnerabilityCollection(id=str(directory), result=result)
        return output_handler.process_scan_diff_result(scan)
    except NotAMergeRequestError:
        click.echo(
            (
                "WARNING: scan ci expects to be run in a merge-request pipeline.\n"
                "No target branch could be identified, will perform a scan all instead.\n"
                "This is a fallback behaviour, that will be removed in a future version."
            ),
            err=True,
        )
        result = sca_scan_all(ctx, directory=directory, scan_mode=ScanMode.CI_ALL)
        scan = SCAScanAllVulnerabilityCollection(id=str(directory), result=result)
        return output_handler.process_scan_all_result(scan)



from pathlib import Path
from typing import Any, Optional, Sequence, Union

import click
from pygitguardian.iac_models import IaCDiffScanResult, IaCScanParameters

from ggshield.cmd.iac.scan.iac_scan_common_options import (
    add_iac_scan_common_options,
    update_context,
)
from ggshield.cmd.iac.scan.iac_scan_utils import (
    IaCSkipScanResult,
    augment_unignored_issues,
    create_output_handler,
    filter_iac_filepaths,
    get_git_filepaths,
    get_iac_tar,
    handle_scan_error,
)
from ggshield.cmd.utils.common_decorators import display_beta_warning, exception_wrapper
from ggshield.cmd.utils.common_options import (
    directory_argument,
    reference_option,
    staged_option,
)
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.cmd.utils.files import check_directory_not_ignored
from ggshield.core.git_hooks.ci.supported_ci import SupportedCI
from ggshield.core.scan import ScanContext, ScanMode
from ggshield.core.tar_utils import INDEX_REF, get_empty_tar
from ggshield.core.text_utils import display_info
from ggshield.utils.files import is_path_excluded
from ggshield.utils.git_shell import (
    Filemode,
    get_diff_files_status,
    get_filepaths_from_ref,
)
from ggshield.verticals.iac.collection.iac_diff_scan_collection import (
    IaCDiffScanCollection,
)
from ggshield.verticals.iac.filter import is_iac_file_path


@click.command()
@add_iac_scan_common_options()
@reference_option
@staged_option
@directory_argument
@click.pass_context
@display_beta_warning
@exception_wrapper
def scan_diff_cmd(
    ctx: click.Context,
    exit_zero: bool,
    minimum_severity: str,
    ignore_policies: Sequence[str],
    ignore_paths: Sequence[str],
    ref: str,
    staged: bool,
    directory: Optional[Path],
    **kwargs: Any,
) -> int:
    """
    Scan a Git repository for changes in IaC vulnerabilities between two states.

    The scan is successful if no *new* IaC vulnerability was found.

    By default, the output will show:
    - The number of known IaC vulnerabilities resolved by the changes
    - The number of known IaC vulnerabilities left untouched
    - The number and the list of new IaC vulnerabilities introduced by the changes
    """
    if directory is None:
        directory = Path().resolve()
    update_context(ctx, exit_zero, minimum_severity, ignore_policies, ignore_paths)

    result = iac_scan_diff(
        ctx, directory, ref, staged, scan_mode=ScanMode.DIRECTORY_DIFF
    )
    ctx_obj = ContextObj.get(ctx)
    augment_unignored_issues(ctx_obj.config.user_config, result)
    return display_iac_scan_diff_result(ctx, directory, result)


def iac_scan_diff(
    ctx: click.Context,
    directory: Path,
    previous_ref: Optional[str],
    include_staged: bool,
    scan_mode: ScanMode,
    current_ref: Optional[str] = None,
    ci_mode: Optional[SupportedCI] = None,
) -> Union[IaCDiffScanResult, IaCSkipScanResult, None]:
    """
    Performs a diff scan for IaC vulnerabilities,
    comparing two git reference. Vulnerabilities are flagged as new, removed or
    remaining depending on whether they appear in the `current_ref` and `previous_ref`
    git references.

    :param ctx: click.Context with CLI arguments
    :param directory: path to the location we want to scan.
    :param previous_ref: git reference to the state of reference for the analysis
    :param include_staged: bool whether or not we want to consider the staged files
    only when the current reference is set to None.
    :param scan_mode: a string describing the type of scan, for API analytics.
    :param current_ref: optional git reference to the current state, defaults to None.
    When set to None, the current state is the indexed files currently on disk.
    :return: IacDiffScanResult if the scan was performed; IaCSkipScanResult if the scan
    was skipped (i.e. no IaC files were detected or changed between the two references)
    """
    ctx_obj = ContextObj.get(ctx)
    config = ctx_obj.config
    client = ctx_obj.client
    exclusion_regexes = ctx_obj.exclusion_regexes

    check_directory_not_ignored(directory, exclusion_regexes)

    verbose = config.user_config.verbose if config and config.user_config else False
    if verbose:
        if previous_ref is None:
            display_info("> No file to scan in reference.")
        else:
            display_info(f"> Scanned files in reference {previous_ref}")
            filepaths = filter_iac_filepaths(
                directory, get_git_filepaths(directory, previous_ref)
            )
            for filepath in filepaths:
                display_info(f"- {click.format_filename(filepath)}")
            display_info("")

    if current_ref is None:
        current_ref = INDEX_REF if include_staged else "HEAD"
    if verbose:
        if include_staged:
            display_info("> Scanned files in current state (staged)")
        else:
            display_info("> Scanned files in current state")
        filepaths = filter_iac_filepaths(
            directory, get_git_filepaths(directory=directory, ref=current_ref)
        )
        for filepath in filepaths:
            display_info(f"- {click.format_filename(filepath)}")

    modified_iac_files = []

    # Check if IaC files were created, deleted or modified
    if previous_ref is None:
        # This means we are scanning all commits up to now.
        filepaths = get_filepaths_from_ref(wd=str(directory), ref=current_ref)
        modified_iac_files = list(filter(is_iac_file_path, filepaths))
    else:
        files_status = get_diff_files_status(
            wd=str(directory),
            current_ref=current_ref,
            ref=previous_ref,
            staged=include_staged,
            similarity=100,
        )
        modified_modes = [Filemode.NEW, Filemode.DELETE, Filemode.MODIFY]
        modified_iac_files = [
            file
            for file, mode in files_status.items()
            if mode in modified_modes
            and not is_path_excluded(file, exclusion_regexes)
            and is_iac_file_path(file)
        ]

    if len(modified_iac_files) == 0:
        return IaCSkipScanResult()

    reference_tar = (
        get_iac_tar(directory, previous_ref, exclusion_regexes)
        if previous_ref is not None
        else get_empty_tar()
    )
    current_tar = get_iac_tar(directory, current_ref, exclusion_regexes)

    scan_parameters = IaCScanParameters(
        list({ignored.policy for ignored in config.user_config.iac.ignored_policies}),
        config.user_config.iac.minimum_severity,
    )

    scan = client.iac_diff_scan(
        reference_tar,
        current_tar,
        scan_parameters,
        ScanContext(
            command_path=ctx.command_path,
            scan_mode=(
                scan_mode if ci_mode is None else f"{scan_mode.value}/{ci_mode.value}"
            ),
            extra_headers={"Ci-Mode": str(ci_mode.value)} if ci_mode else None,
            target_path=directory,
        ).get_http_headers(),
    )

    if not isinstance(scan, IaCDiffScanResult):
        handle_scan_error(client, scan)
        return None
    return scan


def display_iac_scan_diff_result(
    ctx: click.Context,
    directory: Path,
    result: Union[IaCDiffScanResult, IaCSkipScanResult, None],
) -> int:
    output_handler = create_output_handler(ctx)
    if isinstance(result, IaCSkipScanResult):
        return output_handler.process_skip_diff_scan()
    scan = IaCDiffScanCollection(id=str(directory), result=result)
    return output_handler.process_diff_scan(scan)



import logging
from pathlib import Path
from typing import Any, Sequence

import click
from pygitguardian.iac_models import IaCScanResult

from ggshield.cmd.iac.scan.diff import iac_scan_diff
from ggshield.cmd.iac.scan.iac_scan_common_options import (
    add_iac_scan_common_options,
    update_context,
)
from ggshield.cmd.iac.scan.iac_scan_utils import (
    IaCSkipScanResult,
    augment_unignored_issues,
    create_output_handler,
)
from ggshield.cmd.utils.common_decorators import display_beta_warning, exception_wrapper
from ggshield.cmd.utils.common_options import all_option
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.core.git_hooks.prereceive import get_breakglass_option, parse_stdin
from ggshield.core.scan.scan_mode import ScanMode
from ggshield.utils.git_shell import EMPTY_TREE, check_git_ref, is_valid_git_commit_ref
from ggshield.verticals.iac.collection.iac_diff_scan_collection import (
    IaCDiffScanCollection,
)
from ggshield.verticals.iac.collection.iac_path_scan_collection import (
    IaCPathScanCollection,
)


logger = logging.getLogger(__name__)

REMEDIATION_MESSAGE = """  A pre-receive hook set server side prevents you from pushing IaC vulnerabilities.
Apply the recommended remediation steps to remove the vulnerability."""


@click.command()
@add_iac_scan_common_options()
@all_option
@click.pass_context
@display_beta_warning
@exception_wrapper
def scan_pre_receive_cmd(
    ctx: click.Context,
    scan_all: bool,
    exit_zero: bool,
    minimum_severity: str,
    ignore_policies: Sequence[str],
    ignore_paths: Sequence[str],
    **kwargs: Any,
) -> int:
    """
    Scan a Git repository for changes in IaC vulnerabilities in the received pushed commits.
    This is intended to be used as a pre-receive hook.

    The scan is successful if no *new* IaC vulnerability was found, unless `--all` is used,
    in which case the scan is only successful if no IaC vulnerability (old and new) was found.

    By default, the output will show:
    - The number of known IaC vulnerabilities resolved by the changes
    - The number of known IaC vulnerabilities left untouched
    - The number and the list of new IaC vulnerabilities introduced by the changes

    It is the remote equivalent of the `iac scan pre-push` command.

    Note that it is not currently possible to scan a specific sub-directory of the repo.
    """
    if get_breakglass_option():
        return 0

    before_after = parse_stdin()
    if before_after is None:
        return 0
    else:
        before, after = before_after

    update_context(ctx, exit_zero, minimum_severity, ignore_policies, ignore_paths)

    if scan_all:
        # In the pre-receive context, we do not have access to the files,
        # only git objects. Instead of doing an `iac scan all` command,
        # we perform a diff scan with the root of the git tree.
        # Output is handled afterwards, as a scan all.
        before = EMPTY_TREE

    current_path = Path().resolve()

    check_git_ref(wd=str(current_path), ref=after)
    if not is_valid_git_commit_ref(wd=str(current_path), ref=before):
        # When we have a single non-empty commit in the tree,
        # `before` is set to the parent commit which does not exist
        before = None

    result = iac_scan_diff(
        ctx=ctx,
        directory=current_path,
        previous_ref=before,
        include_staged=False,
        current_ref=after,
        scan_mode=ScanMode.PRE_RECEIVE_ALL if scan_all else ScanMode.PRE_RECEIVE_DIFF,
    )
    augment_unignored_issues(ContextObj.get(ctx).config.user_config, result)

    output_handler = create_output_handler(ctx)

    if isinstance(result, IaCSkipScanResult):
        return output_handler.process_skip_diff_scan()

    scan = IaCDiffScanCollection(id=str(current_path), result=result)

    if result is not None:
        if scan_all:
            # If we performed a scan all, we can convert the diff scan result to
            # a path scan, extracting the new vulnerabilities.
            result_all = IaCScanResult(
                id=result.id,
                iac_engine_version=result.iac_engine_version,
                entities_with_incidents=result.entities_with_incidents.new,
            )
            result_all.status_code = result.status_code
            scan_all_collection = IaCPathScanCollection(
                id=str(current_path), result=result_all
            )
            return output_handler.process_scan(scan_all_collection)

    return output_handler.process_diff_scan(scan)



from datetime import datetime, timezone
from pathlib import Path
from typing import List, Optional, Pattern, Set, Tuple, Type, Union

import click
from pygitguardian.client import GGClient, _create_tar
from pygitguardian.sca_models import (
    ComputeSCAFilesResult,
    SCAIgnoredVulnerability,
    SCAScanAllOutput,
    SCAScanDiffOutput,
    SCAScanParameters,
)

from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.cmd.utils.files import check_directory_not_ignored
from ggshield.core.config.user_config import SCAConfig
from ggshield.core.dirs import get_project_root_dir
from ggshield.core.errors import APIKeyCheckError, UnexpectedError
from ggshield.core.scan.scan_context import ScanContext
from ggshield.core.scan.scan_mode import ScanMode
from ggshield.core.tar_utils import INDEX_REF, get_empty_tar, tar_from_ref_and_filepaths
from ggshield.core.text_utils import display_error, display_info
from ggshield.verticals.sca.file_selection import (
    get_all_files_from_sca_paths,
    sca_files_from_git_repo,
)
from ggshield.verticals.sca.output.handler import SCAOutputHandler
from ggshield.verticals.sca.output.json_handler import SCAJsonOutputHandler
from ggshield.verticals.sca.output.text_handler import SCATextOutputHandler


def get_scan_params_from_config(sca_config: SCAConfig) -> SCAScanParameters:
    return SCAScanParameters(
        minimum_severity=sca_config.minimum_severity,
        ignored_vulnerabilities=[
            SCAIgnoredVulnerability(
                identifier=ignored_vuln.identifier, path=ignored_vuln.path
            )
            for ignored_vuln in sca_config.ignored_vulnerabilities
            if ignored_vuln.until is None
            or ignored_vuln.until >= datetime.now(tz=timezone.utc)
        ],
        ignore_fixable=sca_config.ignore_fixable,
        ignore_not_fixable=sca_config.ignore_not_fixable,
    )


def sca_scan_all(
    ctx: click.Context, directory: Path, scan_mode: ScanMode = ScanMode.DIRECTORY
) -> SCAScanAllOutput:
    """
    Scan an entire directory for SCA Vulnerabilities.

    - List SCA related files with a first call to SCA compute files API
    - Create a tar archive with the required files contents
    - Launches the scan with a call to SCA public API
    """
    ctx_obj = ContextObj.get(ctx)
    config = ctx_obj.config
    client = ctx_obj.client
    exclusion_regexes = ctx_obj.exclusion_regexes

    check_directory_not_ignored(directory, exclusion_regexes)

    sca_filepaths, sca_filter_status_code = get_sca_scan_all_filepaths(
        directory=directory,
        exclusion_regexes=exclusion_regexes,
        verbose=config.user_config.verbose,
        client=client,
    )

    if len(sca_filepaths) == 0:
        display_info("No file to scan.")
        # Not an error, return an empty SCAScanAllOutput
        # with the status code returned by first call
        empty_output = SCAScanAllOutput(scanned_files=[], found_package_vulns=[])
        empty_output.status_code = sca_filter_status_code
        return empty_output

    root = get_project_root_dir(directory)
    relative_paths = [
        str((directory / x).resolve().relative_to(root)) for x in sca_filepaths
    ]

    scan_parameters = get_scan_params_from_config(config.user_config.sca)

    tar = _create_tar(root, relative_paths)

    # Call to full scan API and get results
    scan_result = client.sca_scan_directory(
        tar,
        scan_parameters,
        ScanContext(
            command_path=ctx.command_path,
            scan_mode=scan_mode,
            target_path=directory,
        ).get_http_headers(),
    )

    if not isinstance(scan_result, SCAScanAllOutput):
        if scan_result.status_code == 401:
            raise APIKeyCheckError(client.base_uri, "Invalid API key.")
        display_error("Error scanning.")
        display_error(str(scan_result))
        raise UnexpectedError("Unexpected error while performing SCA scan all.")

    return scan_result


def get_sca_scan_all_filepaths(
    directory: Path,
    exclusion_regexes: Set[Pattern[str]],
    verbose: bool,
    client: GGClient,
) -> Tuple[List[str], int]:
    """
    Retrieve SCA related files of a directory.
    First get all filenames that are not in blacklisted directories, then calls SCA compute files
    API to filter SCA related files.
    """
    all_filepaths = get_all_files_from_sca_paths(
        path=directory,
        exclusion_regexes=exclusion_regexes,
        verbose=verbose,
        # If the repository is a git repository, ignore untracked files
        ignore_git=True,
    )

    # API Call to filter SCA files
    response = client.compute_sca_files(files=all_filepaths)

    # First check is required by pyright to know that status_code cannot be None
    if response.status_code is None or not isinstance(response, ComputeSCAFilesResult):
        if response.status_code == 401:
            raise APIKeyCheckError(client.base_uri, "Invalid API key.")
        display_error("Error while filtering SCA related files.")
        display_error(str(response))
        raise UnexpectedError("Unexpected error while filtering SCA related files.")

    # Only sca_files field is useful in the case of a full_scan,
    # all the potential files already exist in `all_filepaths`
    sca_files = response.sca_files
    if verbose:
        display_info("> Scanned files:")
        for filename in sca_files:
            display_info(f"- {click.format_filename(filename)}")

    return sca_files, response.status_code


def create_output_handler(ctx: click.Context) -> SCAOutputHandler:
    """Read objects defined in ctx.obj and create the appropriate OutputHandler
    instance"""
    ctx_obj = ContextObj.get(ctx)
    output_handler_cls: Type[SCAOutputHandler]
    if ctx_obj.use_json:
        output_handler_cls = SCAJsonOutputHandler
    else:
        output_handler_cls = SCATextOutputHandler
    config = ctx_obj.config
    return output_handler_cls(
        verbose=config.user_config.verbose, exit_zero=config.user_config.exit_zero
    )


def sca_scan_diff(
    ctx: click.Context,
    directory: Path,
    previous_ref: Optional[str],
    scan_mode: Union[ScanMode, str],
    ci_mode: Optional[str] = None,
    include_staged: bool = False,
    current_ref: Optional[str] = None,
) -> SCAScanDiffOutput:
    """
    Performs a diff scan for SCA vulnerabilities,
    comparing two git reference. Vulnerabilities are flagged as new, removed or
    remaining depending on whether they appear in the `current_ref` and `previous_ref`
    git references.

    :param ctx: click.Context with CLI arguments
    :param directory: path to the location we want to scan.
    :param previous_ref: git reference to the state of reference for the analysis
    :param include_staged: bool whether or not we want to consider the staged files
    only when the current reference is set to None.
    :param current_ref: optional git reference to the current state, defaults to None.
    When set to None, the current state is the indexed files currently on disk.
    :return: SCAScanDiffOutput object.
    """
    ctx_obj = ContextObj.get(ctx)
    config = ctx_obj.config
    client = ctx_obj.client
    exclusion_regexes = ctx_obj.exclusion_regexes

    check_directory_not_ignored(directory, exclusion_regexes)

    if current_ref is None:
        current_ref = INDEX_REF if include_staged else "HEAD"
    if current_ref == previous_ref:
        display_info("SCA scan diff comparing identical versions, scan skipped.")
        return SCAScanDiffOutput(scanned_files=[], added_vulns=[], removed_vulns=[])

    if previous_ref is None:
        previous_files = []
    else:
        previous_files = sca_files_from_git_repo(
            directory,
            previous_ref,
            client,
            exclusion_regexes=exclusion_regexes,
            verbose=config.user_config.verbose,
        )

    current_files = sca_files_from_git_repo(
        directory,
        current_ref,
        client,
        exclusion_regexes=exclusion_regexes,
        verbose=config.user_config.verbose,
    )

    if len(previous_files) == 0 and len(current_files) == 0:
        display_info("No file to scan.")
        return SCAScanDiffOutput(scanned_files=[], added_vulns=[], removed_vulns=[])

    if previous_ref is None:
        previous_tar = get_empty_tar()
    else:
        previous_tar = tar_from_ref_and_filepaths(
            ref=previous_ref, filepaths=previous_files, wd=str(directory)
        )

    current_tar = tar_from_ref_and_filepaths(
        ref=current_ref, filepaths=current_files, wd=str(directory)
    )

    scan_parameters = get_scan_params_from_config(config.user_config.sca)

    response = client.scan_diff(
        reference=previous_tar,
        current=current_tar,
        scan_parameters=scan_parameters,
        extra_headers=ScanContext(
            command_path=ctx.command_path,
            scan_mode=scan_mode,
            target_path=directory,
            extra_headers={"Ci-Mode": ci_mode} if ci_mode else None,
        ).get_http_headers(),
    )

    if not isinstance(response, SCAScanDiffOutput):
        if response.status_code == 401:
            raise APIKeyCheckError(client.base_uri, "Invalid API key.")
        display_error("Error while scanning diff.")
        display_error(str(response))
        raise UnexpectedError("Unexpected error while scanning diff.")

    return response



import shutil
from collections import Counter, defaultdict
from datetime import datetime
from io import StringIO
from pathlib import Path
from typing import (
    Any,
    ClassVar,
    DefaultDict,
    Dict,
    Generator,
    List,
    NamedTuple,
    Optional,
)

from pygitguardian.iac_models import (
    IaCDiffScanEntities,
    IaCFileResult,
    IaCVulnerability,
)

from ggshield.core.dirs import get_project_root_dir
from ggshield.core.lines import Line, get_lines_from_content, get_offset, get_padding
from ggshield.core.scan import File
from ggshield.core.text_utils import (
    STYLE,
    clip_long_line,
    file_diff_info,
    file_info,
    format_text,
    pluralize,
)
from ggshield.utils.git_shell import Filemode
from ggshield.verticals.iac.collection.iac_diff_scan_collection import (
    IaCDiffScanCollection,
)
from ggshield.verticals.iac.collection.iac_path_scan_collection import (
    IaCPathScanCollection,
)
from ggshield.verticals.iac.output.iac_output_handler import IaCOutputHandler


class GroupedIncidents(NamedTuple):
    name: str
    new: List[IaCFileResult]
    unchanged: List[IaCFileResult]
    deleted: List[IaCFileResult]


def group_incidents_by_filename(
    incidents: IaCDiffScanEntities,
) -> Generator[GroupedIncidents, None, None]:
    filenames = []
    statuses: Dict[str, DefaultDict] = {
        "new": defaultdict(list),
        "unchanged": defaultdict(list),
        "deleted": defaultdict(list),
    }
    for status in statuses:
        for entry in getattr(incidents, status):
            filename = entry.filename
            if filename not in filenames:
                filenames.append(filename)
            statuses[status].setdefault(filename, [])
            statuses[status][filename].append(entry)
    for filename in filenames:
        new: List[IaCFileResult] = statuses.get("new", {}).get(filename, [])
        unchanged: List[IaCFileResult] = statuses.get("unchanged", {}).get(filename, [])
        deleted: List[IaCFileResult] = statuses.get("deleted", {}).get(filename, [])
        yield GroupedIncidents(filename, new, unchanged, deleted)


class IaCTextOutputHandler(IaCOutputHandler):
    nb_lines: ClassVar[int] = 3

    def _process_scan_impl(self, scan: IaCPathScanCollection) -> str:
        scan_buf = StringIO()

        entities_without_ignored = scan.get_entities_without_ignored()
        if scan.result is not None and isinstance(entities_without_ignored, List):
            scan_buf.write(iac_engine_version(scan.result.iac_engine_version))
            # List incidents if any
            source_basedir = get_project_root_dir(Path(scan.id))
            for file_result in entities_without_ignored:
                scan_buf.write(
                    self.process_iac_file_result(
                        source_basedir / file_result.filename, file_result
                    )
                )
            # Show no incidents if none
            if len(entities_without_ignored) == 0:
                scan_buf.write(no_iac_vulnerabilities())
        return scan_buf.getvalue()

    def _process_diff_scan_impl_not_verbose(self, scan: IaCDiffScanCollection) -> str:
        scan_buf = StringIO()

        entities_without_ignored = scan.get_entities_without_ignored()
        if scan.result is not None and isinstance(
            entities_without_ignored, IaCDiffScanEntities
        ):
            # Add iac version on output
            scan_buf.write(iac_engine_version(scan.result.iac_engine_version))
            # Show no incidents if none
            if len(entities_without_ignored.new) == 0:
                scan_buf.write(
                    format_text(
                        "\nNo new incidents have been found\n", STYLE["no_secret"]
                    )
                )
            else:
                source_basedir = get_project_root_dir(Path(scan.id))
                for filename, new, _, _ in group_incidents_by_filename(
                    entities_without_ignored
                ):
                    if not new:
                        continue
                    num_new = sum(len(e.incidents) for e in new)
                    scan_buf.write(file_diff_info(filename, num_new, None, None))
                    # List new incidents if any
                    for file_result in new:
                        scan_buf.write(
                            self.process_iac_diff_result(
                                source_basedir / file_result.filename, file_result
                            )
                        )
            # Show summary
            scan_buf.write(
                diff_scan_summary(
                    entities_without_ignored.new,
                    entities_without_ignored.unchanged,
                    entities_without_ignored.deleted,
                )
            )
        return scan_buf.getvalue()

    def _process_diff_scan_impl_verbose(self, scan: IaCDiffScanCollection) -> str:
        entities_without_ignored = scan.get_entities_without_ignored()
        if scan.result is None or not isinstance(
            entities_without_ignored, IaCDiffScanEntities
        ):
            return ""

        scan_buf = StringIO()
        # Add iac version on output
        scan_buf.write(iac_engine_version(scan.result.iac_engine_version))
        # Show no incidents if none
        num_new = sum(len(e.incidents) for e in entities_without_ignored.new)
        num_unchanged = sum(
            len(e.incidents) for e in entities_without_ignored.unchanged
        )
        num_deleted = sum(len(e.incidents) for e in entities_without_ignored.deleted)
        total_vulns_count = num_new + num_unchanged + num_deleted
        source_basedir = get_project_root_dir(Path(scan.id))
        if total_vulns_count == 0:
            scan_buf.write(no_iac_vulnerabilities())
        else:
            for filename, new, unchanged, deleted in group_incidents_by_filename(
                entities_without_ignored
            ):
                num_new = sum(len(e.incidents) for e in new)
                num_unchanged = sum(len(e.incidents) for e in unchanged)
                num_deleted = sum(len(e.incidents) for e in deleted)
                scan_buf.write(
                    file_diff_info(filename, num_new, num_unchanged, num_deleted)
                )

                # List deleted incidents if any
                for file_result in deleted:
                    scan_buf.write(
                        self.process_iac_diff_result(
                            source_basedir / file_result.filename,
                            file_result,
                            "REMOVED",
                        )
                    )
                # List unchanged incidents if any
                for file_result in unchanged:
                    scan_buf.write(
                        self.process_iac_diff_result(
                            source_basedir / file_result.filename,
                            file_result,
                            "PERSISTING",
                        )
                    )
                # List new incidents if any
                for file_result in new:
                    scan_buf.write(
                        self.process_iac_diff_result(
                            source_basedir / file_result.filename, file_result, "NEW"
                        )
                    )
        # Show summary
        scan_buf.write(
            diff_scan_summary(
                entities_without_ignored.new,
                entities_without_ignored.unchanged,
                entities_without_ignored.deleted,
            )
        )
        return scan_buf.getvalue()

    def _process_diff_scan_impl(self, scan: IaCDiffScanCollection) -> str:
        if self.verbose:
            return self._process_diff_scan_impl_verbose(scan)
        return self._process_diff_scan_impl_not_verbose(scan)

    def _process_skip_scan_impl(self) -> str:
        return "> No IaC files detected. Skipping."

    def _process_skip_diff_scan_impl(self) -> str:
        return "> No IaC files changed. Skipping."

    def process_iac_file_result(
        self, file_path: Path, file_result: IaCFileResult, prefix: Optional[str] = None
    ) -> str:
        """
        Build readable message on the found incidents for a specific file

        :param file_path: The full path to the file, used to read the content of the file
        :param file_result: The file results from the IaC scanning API
        :return: The formatted message to display
        """
        result_buf = StringIO()

        result_buf.write(file_info(file_result.filename, len(file_result.incidents)))

        try:
            file = File(file_path)
            lines: List[Line] = get_lines_from_content(file.content, Filemode.FILE)
        except Exception:
            lines = []

        for issue_n, vulnerability in enumerate(file_result.incidents, 1):
            result_buf.write(
                iac_vulnerability_header(issue_n, vulnerability, prefix=prefix)
            )
            result_buf.write(
                iac_vulnerability_documentation(vulnerability.documentation_url)
            )
            result_buf.write(iac_vulnerability_severity_line(vulnerability.severity))
            if len(lines) == 0:
                result_buf.write(
                    iac_vulnerability_location_failed(
                        vulnerability.line_start, vulnerability.line_end
                    )
                )
            else:
                result_buf.write(
                    iac_vulnerability_location(
                        lines,
                        vulnerability.line_start,
                        vulnerability.line_end,
                        self.nb_lines,
                        clip_long_lines=not self.verbose,
                    )
                )
            if vulnerability.ignored_until is not None:
                result_buf.write(
                    iac_vulnerability_end_of_ignored_period(vulnerability.ignored_until)
                )

        return result_buf.getvalue()

    def process_iac_diff_result(
        self, file_path: Path, file_result: IaCFileResult, prefix: Optional[str] = None
    ) -> str:
        """
        Build readable message on the found incidents for a specific file

        :param file_path: The full path to the file, used to read the content of the file
        :param file_result: The file results from the IaC scanning API
        :return: The formatted message to display
        """
        result_buf = StringIO()

        try:
            file = File(file_path)
            lines: List[Line] = get_lines_from_content(file.content, Filemode.FILE)
        except Exception:
            lines = []

        for issue_n, vulnerability in enumerate(file_result.incidents, 1):
            result_buf.write(
                iac_vulnerability_header(issue_n, vulnerability, prefix=prefix)
            )
            result_buf.write(
                iac_vulnerability_documentation(vulnerability.documentation_url)
            )
            result_buf.write(iac_vulnerability_severity_line(vulnerability.severity))
            if len(lines) == 0:
                result_buf.write(
                    iac_vulnerability_location_failed(
                        vulnerability.line_start, vulnerability.line_end
                    )
                )
            else:
                result_buf.write(
                    iac_vulnerability_location(
                        lines,
                        vulnerability.line_start,
                        vulnerability.line_end,
                        self.nb_lines,
                        clip_long_lines=not self.verbose,
                    )
                )
            if vulnerability.ignored_until is not None:
                result_buf.write(
                    iac_vulnerability_end_of_ignored_period(vulnerability.ignored_until)
                )

        return result_buf.getvalue()


def iac_vulnerability_header(
    issue_n: int, vulnerability: IaCVulnerability, prefix: Optional[str] = None
) -> str:
    """
    Build a header for the iac policy break.
    """
    return "\n{}{} Incident {} ({}): {}: {} ({})\n".format(
        format_text(">>>", STYLE["detector_line_start"]),
        "" if prefix is None else f" {prefix}:",
        issue_n,
        format_text("IaC", STYLE["detector"]),
        format_text(vulnerability.component, STYLE["detector"]),
        format_text(vulnerability.policy, STYLE["policy"]),
        format_text(vulnerability.policy_id, STYLE["policy"]),
    )


def iac_vulnerability_documentation(doc_url: str) -> str:
    return f"More at {doc_url}\n"


def iac_vulnerability_severity_line(severity: str) -> str:
    """
    Build a line to output the severity of a vulnerability
    """
    if severity == "CRITICAL":
        severity_string = "Critical"
        style = STYLE["iac_vulnerability_critical"]
    elif severity == "HIGH":
        severity_string = "High"
        style = STYLE["iac_vulnerability_high"]
    elif severity == "MEDIUM":
        severity_string = "Medium"
        style = STYLE["iac_vulnerability_medium"]
    elif severity == "LOW":
        severity_string = "Low"
        style = STYLE["iac_vulnerability_low"]
    else:  # In the other case, print `severity``
        severity_string = severity
        style = STYLE["iac_vulnerability_unknown"]

    return f"Severity: {format_text(severity_string, style)}\n"


def iac_vulnerability_location(
    lines: List[Line],
    line_start: int,
    line_end: int,
    nb_lines: int,
    clip_long_lines: bool = False,
) -> str:
    msg = StringIO()
    padding = get_padding(lines)
    offset = get_offset(padding)
    max_width = shutil.get_terminal_size()[0] - offset if clip_long_lines else 0
    for line_nb in range(
        max(0, line_start - nb_lines), min(len(lines) - 1, line_end + nb_lines)
    ):
        msg.write(
            lines[line_nb].build_line_count(
                padding, line_start - 1 <= line_nb <= line_end - 1
            )
        )
        line_content = lines[line_nb].content

        if max_width:
            line_content = clip_long_line(line_content, max_width, after=True)
        msg.write(f"{line_content}\n")
    return msg.getvalue()


def iac_vulnerability_location_failed(
    line_start: int,
    line_end: int,
) -> str:
    return f"\nThe incident was found between lines {line_start} and {line_end}\n"  # noqa: E501


def iac_vulnerability_end_of_ignored_period(
    ignored_until: datetime,
) -> str:
    return f"\nThe incident is no longer ignored in the scan since {ignored_until.strftime('%Y-%m-%d')}\n"  # noqa: E501


def iac_engine_version(iac_engine_version: str) -> str:
    return f"\niac-engine-version: {iac_engine_version}\n"


def no_iac_vulnerabilities() -> str:
    """
    Build a message if no IaC vulnerabilities were found.
    """
    return format_text("\nNo incidents have been found\n", STYLE["no_secret"])


def diff_scan_summary(
    new: List[IaCFileResult],
    unchanged: List[IaCFileResult],
    deleted: List[IaCFileResult],
) -> str:
    def detail(entries: List[IaCFileResult]) -> str:
        def _get_style(severity: str) -> Dict[str, Dict[str, Any]]:
            return STYLE.get(
                f"iac_vulnerability_{severity.lower()}",
                STYLE["iac_vulnerability_unknown"],
            )

        count = Counter(
            incident.severity for entry in entries for incident in entry.incidents
        )
        severity_order = {
            k: i for i, k in enumerate(["LOW", "MEDIUM", "HIGH", "CRITICAL"])
        }
        formatted_count = [
            format_text(f"{key}: {val}", _get_style(key))
            for key, val in sorted(
                count.items(), key=lambda x: severity_order.get(x[0], -1)
            )
        ]
        if len(formatted_count) == 0:
            return ""
        return f" ({', '.join(formatted_count)})"

    def label_incident(n: int) -> str:
        return pluralize("incident", n, "incidents")

    num_deleted = sum(len(entry.incidents) for entry in deleted)
    num_unchanged = sum(len(entry.incidents) for entry in unchanged)
    num_new = sum(len(entry.incidents) for entry in new)

    buf = StringIO()
    buf.write("\nSummary of changes:\n")
    buf.write(
        format_text(
            f"[-] {num_deleted} {label_incident(num_deleted)} deleted",
            STYLE[
                "iac_deleted_vulnerability" if num_deleted > 0 else "iac_dim_summary"
            ],
        )
    )
    buf.write(f"{detail(deleted)}\n")
    buf.write(
        format_text(
            f"[~] {num_unchanged} {label_incident(num_unchanged)} remaining",
            STYLE[
                (
                    "iac_remaining_vulnerability"
                    if num_unchanged > 0
                    else "iac_dim_summary"
                )
            ],
        )
    )
    buf.write(f"{detail(unchanged)}\n")
    buf.write(
        format_text(
            f"[+] {num_new} new {label_incident(num_new)} detected",
            STYLE["iac_new_vulnerability" if num_new > 0 else "iac_dim_summary"],
        )
    )
    buf.write(f"{detail(new)}\n")
    return buf.getvalue()



import functools
import json
import logging
import os
import urllib.parse as urlparse
import webbrowser
from base64 import urlsafe_b64encode
from datetime import datetime
from hashlib import sha256
from http.server import BaseHTTPRequestHandler, HTTPServer
from typing import Any, Dict, List, Optional

import click
from oauthlib.oauth2 import OAuth2Error, WebApplicationClient

from ggshield.core.client import (
    check_client_api_key,
    create_client,
    create_client_from_config,
    create_session,
)
from ggshield.core.config import Config, InstanceConfig
from ggshield.core.errors import APIKeyCheckError, UnexpectedError
from ggshield.core.url_utils import urljoin
from ggshield.utils.datetime import get_pretty_date


CLIENT_ID = "ggshield_oauth"
SCAN_SCOPE = "scan"

# potential port range to be used to run local server
# to handle authorization code callback
# this is the largest band of not commonly occupied ports
# https://stackoverflow.com/questions/10476987/best-tcp-port-number-range-for-internal-applications
USABLE_PORT_RANGE = (29170, 29998)

logger = logging.getLogger(__name__)


def get_error_param(parsed_url: urlparse.ParseResult) -> Optional[str]:
    """
    extract the value of the 'error' url param. If not present, return None.
    """
    params = urlparse.parse_qs(parsed_url.query)
    if "error" in params:
        return params["error"][0]
    return None


class OAuthClient:
    """
    Helper class to handle the OAuth authentication flow
    the logic is divided in 2 steps:
    - open the browser on GitGuardian login screen and run a local server to wait for callback
    - handle the oauth callback to exchange an authorization code against a valid access token
    """

    def __init__(self, config: Config, instance: str) -> None:
        self.config = config
        self.instance = instance
        self._oauth_client = WebApplicationClient(CLIENT_ID)
        self._state = ""  # use the `state` property instead
        self._lifetime: Optional[int] = None
        self._login_path = "auth/login"
        self._extra_scopes = []

        # Fields updated by RequestHandler
        self._request_finished = False
        self._request_error_message: Optional[str] = None

        self._access_token: Optional[str] = None
        # If the PAT expiration date has been enforced to respect the workspace policy
        self._expire_at_downsized: bool = False
        self._port = USABLE_PORT_RANGE[0]
        self.server: Optional[HTTPServer] = None

        self._generate_pkce_pair()

    def oauth_process(
        self,
        token_name: Optional[str] = None,
        lifetime: Optional[int] = None,
        login_path: Optional[str] = None,
        extra_scopes: Optional[List[str]] = None,
    ) -> None:
        """
        Handle the whole oauth process which includes
        - opening the user's webbrowser to GitGuardian login page
        - open a server and wait for the callback processing
        """
        # enable redirection to http://localhost
        os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = str(True)

        if token_name is None:
            token_name = "ggshield token " + datetime.today().strftime("%Y-%m-%d")
        self._token_name = token_name
        if login_path is not None:
            self._login_path = login_path

        if lifetime is None:
            lifetime = self.default_token_lifetime
        self._lifetime = lifetime

        if extra_scopes is not None:
            self._extra_scopes = extra_scopes

        self._prepare_server()
        self._redirect_to_login()
        self._wait_for_callback()
        self._print_login_success()

    def process_callback(self, callback_url: str) -> None:
        """
        This function runs within the request handler do_GET method.
        It takes the url of the callback request as argument and does
        - Extract the authorization code
        - Exchange the code against an access token with GitGuardian's api
        - Validate the new token against GitGuardian's api
        - Save the token in configuration
        Any error during this process will raise a OAuthError
        """
        authorization_code = self._get_code(callback_url)
        self._claim_token(authorization_code)
        token_data = self._validate_access_token()
        self._save_token(token_data)

    def _generate_pkce_pair(self) -> None:
        """
        Generate a code verifier (random string) and its sha encoded version to be used
        for the pkce checking process
        """
        self.code_verifier = self._oauth_client.create_code_verifier(128)
        self.code_challenge = (
            urlsafe_b64encode(sha256(self.code_verifier.encode()).digest())
            .decode()
            .rstrip("=")
        )

    def _redirect_to_login(self) -> None:
        """
        Open the user's browser to the GitGuardian ggshield authentication page
        """
        static_params = {
            "auth_mode": "ggshield_login",
            "utm_source": "cli",
            "utm_medium": "login",
            "utm_campaign": "ggshield",
        }
        request_uri = self._oauth_client.prepare_request_uri(
            uri=urljoin(self.dashboard_url, self._login_path),
            redirect_uri=self.redirect_uri,
            scope=[SCAN_SCOPE, *self._extra_scopes],
            code_challenge=self.code_challenge,
            code_challenge_method="S256",
            state=self.state,
            **static_params,
        )
        click.echo(
            f"Complete the login process at:\n"
            f"  {request_uri}.\n"
            "Opening your web browser now..."
        )
        webbrowser.open_new_tab(request_uri)

    def _prepare_server(self) -> None:
        for port in range(*USABLE_PORT_RANGE):
            try:
                self.server = HTTPServer(
                    # only consider requests from localhost on the predetermined port
                    ("127.0.0.1", port),
                    functools.partial(RequestHandler, self),
                )
                self._port = port
                break
            except OSError:
                continue
        else:
            raise UnexpectedError("Could not find unoccupied port.")

    def _wait_for_callback(self) -> None:
        """
        Wait to receive and process the authorization callback on the local server.
        This actually catches HTTP requests made on the previously opened server.
        The callback processing logic is implementend in the request handler class
        and the `process_callback` method
        """
        try:
            while not self._request_finished:
                # Wait for callback on localserver including an authorization code
                # any matching request will get processed by the request handler and
                # the `process_callback` function
                self.server.handle_request()  # type: ignore
        except KeyboardInterrupt:
            raise click.Abort()

        if self._request_error_message is not None:
            # if no error message is attached, the process is considered successful
            raise UnexpectedError(self._request_error_message)

    def _get_code(self, uri: str) -> str:
        """
        Extract the authorization from the incoming request uri and verify that the state from
        the uri match the one stored internally.
        if no code can be extracted or the state is invalid, raise an OAuthError
        else return the extracted code
        """
        try:
            authorization_code = self._oauth_client.parse_request_uri_response(
                uri, self.state
            ).get("code")
        except OAuth2Error:
            authorization_code = None
        if authorization_code is None:
            raise OAuthError("Invalid code or state received from the callback.")
        return authorization_code

    def _claim_token(self, authorization_code: str) -> None:
        """
        Exchange the authorization code with a valid access token using GitGuardian public api.
        If no valid token could be retrieved, exit the authentication process with an error message
        """

        request_params = {"name": self._token_name}
        if self._lifetime is not None:
            request_params["lifetime"] = str(self._lifetime)

        request_body = self._oauth_client.prepare_request_body(
            code=authorization_code,
            redirect_uri=self.redirect_uri,
            code_verifier=self.code_verifier,
            body=urlparse.urlencode(request_params),
        )

        session = create_session(self.config.user_config.allow_self_signed)
        response = session.post(
            urljoin(self.api_url, "/v1/oauth/token"),
            request_body,
            headers={"Content-Type": "application/x-www-form-urlencoded"},
        )

        try:
            dct = response.json()
        except json.decoder.JSONDecodeError:
            raise UnexpectedError(
                f"Server response is not JSON (HTTP code: {response.status_code})."
            )

        if not response.ok:
            detail = dct.get("detail", "[no error message]")
            raise OAuthError(f"Cannot create a token: {detail}.")

        try:
            self._access_token = dct["key"]
        except KeyError:
            raise UnexpectedError("Server did not provide the created token.")

        self._expire_at_downsized = dct.get("expire_at_downsized", False)

    def _validate_access_token(self) -> Dict[str, Any]:
        """
        Validate the token using GitGuardian public api.
        If the token is not valid, exit the authentication process with an error message.
        """
        assert self._access_token is not None
        response = create_client(
            self._access_token,
            self.api_url,
            allow_self_signed=self.config.user_config.allow_self_signed,
        ).get(endpoint="token")
        if not response.ok:
            raise OAuthError("The created token is invalid.")
        return response.json()

    def _save_token(self, api_token_data: Dict[str, Any]) -> None:
        """
        Save the new token in the configuration.
        """
        assert self._access_token is not None
        self.instance_config.init_account(self._access_token, api_token_data)
        self.config.auth_config.save()

    @property
    def instance_config(self) -> InstanceConfig:
        return self.config.auth_config.get_instance(self.instance)

    @property
    def default_token_lifetime(self) -> Optional[int]:
        """
        return the default token lifetime saved in the instance config.
        if None, this will be interpreted as no expiry.
        """
        instance_lifetime = self.instance_config.default_token_lifetime
        if instance_lifetime is not None:
            return instance_lifetime
        return self.config.auth_config.default_token_lifetime

    @property
    def redirect_uri(self) -> str:
        return f"http://localhost:{self._port}"

    @property
    def state(self) -> str:
        """
        Return the state used to verify the auth process.
        The state is included in the redirect_uri and is expected in the callback url.
        Then, if both states don't match, the process fails.
        The state is an url-encoded string dict containing the token name and lifetime
        It is cached to prevent from altering its value during the process
        """
        if not self._state:
            self._state = urlparse.quote(
                json.dumps({"token_name": self._token_name, "lifetime": self._lifetime})
            )
        return self._state

    def check_existing_token(self) -> bool:
        """
        Check if the config already has a non expired token.
        If one could be found, outputs a message including the expiry date
        and return True.
        Else return False
        """
        account = self.instance_config.account
        if account is None or not account.token or self.instance_config.expired:
            return False

        # Check our API key is valid, if not forget it
        client = create_client_from_config(self.config)
        try:
            check_client_api_key(client)
        except APIKeyCheckError:
            # Forget the account
            logger.debug(
                "Account had an API key recorded but it's no longer valid, removing it"
            )
            self.instance_config.account = None
            return False

        message = "ggshield is already authenticated "
        if account.expire_at:
            message += "until " + get_pretty_date(account.expire_at)
        else:
            message += "without an expiry date"
        click.echo(message)
        return True

    def _print_login_success(self) -> None:
        """
        Output the login success message
        """
        assert self.instance_config.account is not None
        expire_at = self.instance_config.account.expire_at

        if expire_at is not None:
            str_date = get_pretty_date(expire_at)
        else:
            str_date = "never"

        expiration_warning = ""
        if self._expire_at_downsized:
            expiration_warning = (
                " Warning: the expiration date has been adjusted to comply with your workspace's"
                " setting for the maximum lifetime of personal access tokens.\n"
            )

        message = (
            "Success! You are now authenticated.\n"
            "The personal access token has been created and stored in your ggshield config.\n\n"
            f"token name: {self._token_name}\n"
            f"token expiration date: {str_date}\n"
            f"{expiration_warning}"
            "\n"
            'You do not need to run "ggshield auth login" again. Future requests will automatically use the token.'
        )

        click.echo(message)

    def get_server_error_message(self, error_code: str) -> str:
        """
        Return the human-readable message associated to the given error code
        """
        if error_code == "too_many_tokens":
            url = urljoin(self.dashboard_url, "/api/personal-access-tokens")
            return (
                "Maximum number of personal access tokens reached. Could not provision a new personal access token.\n"
                f"Go to your workspace to manage your tokens: {url}"
            )
        elif error_code == "invalid_saml":
            return "The given SSO URL is invalid."
        elif error_code == "invalid_scope":
            return "The requested scopes are invalid."
        return f"An unknown server error has occurred (error code: {error_code})."

    @property
    def dashboard_url(self) -> str:
        return self.config.dashboard_url

    @property
    def api_url(self) -> str:
        return self.config.api_url


class RequestHandler(BaseHTTPRequestHandler):
    def __init__(
        self,
        oauth_client: OAuthClient,
        *args: Any,
        **kwargs: Any,
    ):
        # oauth_client must be initialized *before* calling super().__init__(), because
        # BaseHTTPRequestHandler.__init__() calls do_GET().
        self.oauth_client = oauth_client
        super().__init__(*args, **kwargs)

    def do_GET(self) -> None:
        """
        This function process every GET requests received by the server.
        Non-root requests are skipped.
        If an authorization code can be extracted from the URI, attach it to the handler
        so it can be retrieved after the request is processed, then kill the server.
        """
        callback_url: str = self.path
        parsed_url = urlparse.urlparse(callback_url)
        if parsed_url.path != "/":
            self._write_error_response(404, f"Invalid path: {parsed_url.path}")
            return

        self.oauth_client._request_finished = True

        if error_param := get_error_param(parsed_url):
            error_message = self.oauth_client.get_server_error_message(error_param)
            self._handle_error(error_message)
            return

        try:
            self.oauth_client.process_callback(callback_url)
        except OAuthError as error:
            self._handle_error(error.message)
        else:
            self._redirect(
                urljoin(self.oauth_client.dashboard_url, "authenticated"),
            )

    def _write_error_response(self, status_code: int, message: str) -> None:
        """Return a basic HTML error page"""
        self.send_response(status_code)
        self.end_headers()

        content = f"<html><body><h1>Error</h1>{message}</body></html>"
        self.wfile.write(content.encode())

    def _redirect(self, redirect_url: str) -> None:
        self.send_response(301)
        self.send_header("Location", redirect_url)
        self.end_headers()

    def _handle_error(self, error_message: str) -> None:
        self.oauth_client._request_error_message = error_message

        # Redirect to error page
        query = urlparse.urlencode({"message": error_message})
        url = urljoin(self.oauth_client.dashboard_url, f"authentication-error?{query}")
        self._redirect(url)

    def log_message(self, format: str, *args: Any) -> None:
        """Silence log message"""
        return


class OAuthError(Exception):
    """
    Exception raised during the authorization exchange code process
    Its message is caught and will be raised again as a click Exception
    """

    def __init__(self, message: str) -> None:
        self.message = message



import logging
from pathlib import Path
from typing import Any, Sequence

import click
from pygitguardian.sca_models import SCAScanAllOutput

from ggshield.cmd.sca.scan.sca_scan_utils import create_output_handler, sca_scan_diff
from ggshield.cmd.sca.scan.scan_common_options import (
    add_sca_scan_common_options,
    update_context,
)
from ggshield.cmd.utils.common_decorators import exception_wrapper
from ggshield.cmd.utils.common_options import all_option
from ggshield.core.git_hooks.prereceive import get_breakglass_option, parse_stdin
from ggshield.core.scan.scan_mode import ScanMode
from ggshield.utils.git_shell import EMPTY_TREE, check_git_ref, is_valid_git_commit_ref
from ggshield.verticals.sca.collection.collection import (
    SCAScanAllVulnerabilityCollection,
    SCAScanDiffVulnerabilityCollection,
)


logger = logging.getLogger(__name__)

REMEDIATION_MESSAGE = """  A pre-receive hook set server side prevents you from pushing SCA vulnerabilities.
Apply the recommended remediation steps to remove the vulnerability."""


@click.command()
@add_sca_scan_common_options()
@all_option
@click.pass_context
@exception_wrapper
def scan_pre_receive_cmd(
    ctx: click.Context,
    scan_all: bool,
    exit_zero: bool,
    minimum_severity: str,
    ignore_paths: Sequence[str],
    ignore_fixable: bool,
    ignore_not_fixable: bool,
    **kwargs: Any,
) -> int:
    """
    Scans if the received HEAD of a git repository introduces new SCA vulnerabilities.

    This command checks if the current HEAD of a git repository introduces new SCA
    vulnerabilities compared to the remote HEAD of the branch in a pre-receive hook.

    Scanning a repository with this command will not trigger any incident on your dashboard.

    Only metadata such as call time, request size and scan mode is stored server-side.
    """
    if get_breakglass_option():
        return 0

    before_after = parse_stdin()
    if before_after is None:
        return 0
    else:
        before, after = before_after

    update_context(
        ctx,
        exit_zero,
        minimum_severity,
        ignore_paths,
        ignore_fixable,
        ignore_not_fixable,
    )

    if scan_all:
        # In the pre-receive context, we do not have access to the files,
        # only git objects. Instead of doing an `sca scan all` command,
        # we perform a diff scan with the root of the git tree.
        # Output is handled afterwards, as a scan all.
        before = EMPTY_TREE

    current_path = Path().resolve()

    check_git_ref(wd=str(current_path), ref=after)
    if not is_valid_git_commit_ref(wd=str(current_path), ref=before):
        # When we have a single non-empty commit in the tree,
        # `before` is set to the parent commit which does not exist
        before = None

    result = sca_scan_diff(
        ctx=ctx,
        directory=current_path,
        previous_ref=before,
        include_staged=False,
        current_ref=after,
        scan_mode=ScanMode.PRE_RECEIVE_ALL if scan_all else ScanMode.PRE_RECEIVE_DIFF,
    )

    output_handler = create_output_handler(ctx)

    if scan_all:
        # If we performed a scan all, we can convert the diff scan result to
        # a path scan, extracting the new vulnerabilities.
        result_all = SCAScanAllOutput(
            scanned_files=result.scanned_files,
            found_package_vulns=result.added_vulns,
        )
        scan_all_collection = SCAScanAllVulnerabilityCollection(
            id=str(current_path), result=result_all
        )
        return output_handler.process_scan_all_result(scan_all_collection)

    scan = SCAScanDiffVulnerabilityCollection(id=str(current_path), result=result)
    return output_handler.process_scan_diff_result(scan)



from enum import Enum, auto
from pathlib import Path, PurePosixPath
from typing import List, Pattern, Set, Union

from ggshield.utils._binary_extensions import BINARY_EXTENSIONS
from ggshield.utils.git_shell import (
    get_filepaths_from_ref,
    git_ls,
    git_ls_unstaged,
    is_git_dir,
)


class ListFilesMode(Enum):
    """
    Control `get_filepaths()` behavior:

    - FILES_ONLY: list specified paths. Expect them to be plain files, raise an exception if one of them is not.
    - ALL: list all specified paths. If one of the path is a directory, list all its paths recursively.
    - ALL_BUT_GITIGNORED: like ALL, except those ignored by git (listed in .gitignore).
    - GIT_COMMITTED_OR_STAGED: list all committed files and all staged files.
    - GIT_COMMITTED: list only committed files.
    """

    FILES_ONLY = auto()
    GIT_COMMITTED = auto()
    GIT_COMMITTED_OR_STAGED = auto()
    ALL_BUT_GITIGNORED = auto()
    ALL = auto()


class UnexpectedDirectoryError(ValueError):
    """Raise when a directory is used where it is not excepted"""

    def __init__(self, path: Path):
        self.path = path


def is_path_excluded(
    path: Union[str, Path], exclusion_regexes: Set[Pattern[str]]
) -> bool:
    path = Path(path)
    if path.is_dir():
        # The directory exclusion regexes have to end with a slash
        # To check if path is excluded, we need to add a trailing slash
        path_string = f"{PurePosixPath(path)}/"
    else:
        path_string = str(PurePosixPath(path))
    return any(r.search(path_string) for r in exclusion_regexes)


def get_filepaths(
    paths: List[Path],
    exclusion_regexes: Set[Pattern[str]],
    list_files_mode: ListFilesMode,
) -> Set[Path]:
    """
    Retrieve the filepaths from the command.

    :param paths: List of file/dir paths from the command
    :param ignore_git: Ignore that the folder is a git repository
    :raise: UnexpectedDirectoryError if directory is given without --recursive option
    """
    targets: Set[Path] = set()
    for path in paths:
        if path.is_file():
            targets.add(path)
        elif path.is_dir():
            if list_files_mode == ListFilesMode.FILES_ONLY:
                raise UnexpectedDirectoryError(path)

            _targets = set()
            if list_files_mode != ListFilesMode.ALL and is_git_dir(path):
                target_filepaths = (
                    get_filepaths_from_ref("HEAD", wd=path)
                    if list_files_mode == ListFilesMode.GIT_COMMITTED
                    else git_ls(path)
                )
                _targets = {path / x for x in target_filepaths}
                if list_files_mode == ListFilesMode.ALL_BUT_GITIGNORED:
                    _targets.update({path / x for x in git_ls_unstaged(path)})
            else:
                _targets = path.rglob(r"*")

            for file_path in _targets:
                if not is_path_excluded(file_path, exclusion_regexes):
                    targets.add(file_path)
    return targets


def is_path_binary(path: Union[str, Path]) -> bool:
    ext = Path(path).suffix
    # `[1:]` because `ext` starts with a "." but extensions in `BINARY_EXTENSIONS` do not
    return ext[1:] in BINARY_EXTENSIONS



from pathlib import Path
from typing import Any, List

import click

from ggshield.cmd.secret.scan.secret_scan_common_options import (
    add_secret_scan_common_options,
    create_output_handler,
)
from ggshield.cmd.utils.common_decorators import exception_wrapper
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.cmd.utils.files import check_directory_not_ignored
from ggshield.core.scan import ScanContext, ScanMode
from ggshield.core.scan.file import get_files_from_paths
from ggshield.utils.click import RealPath
from ggshield.utils.files import ListFilesMode
from ggshield.verticals.secret import SecretScanCollection, SecretScanner


@click.command()
@click.argument(
    "paths", nargs=-1, type=RealPath(exists=True, resolve_path=True), required=True
)
@click.option("--recursive", "-r", is_flag=True, help="Scan directory recursively.")
@click.option("--yes", "-y", is_flag=True, help="Confirm recursive scan.")
@click.option(
    "--use-gitignore", is_flag=True, help="Honor content of .gitignore files."
)
@add_secret_scan_common_options()
@click.pass_context
@exception_wrapper
def path_cmd(
    ctx: click.Context,
    paths: List[Path],
    recursive: bool,
    yes: bool,
    use_gitignore: bool,
    **kwargs: Any,
) -> int:  # pragma: no cover
    """
    Scan files and directories.
    """
    ctx_obj = ContextObj.get(ctx)
    config = ctx_obj.config
    output_handler = create_output_handler(ctx)
    verbose = config.user_config.verbose

    for path in paths:
        check_directory_not_ignored(path, ctx_obj.exclusion_regexes)

    files = get_files_from_paths(
        paths=paths,
        exclusion_regexes=ctx_obj.exclusion_regexes,
        yes=yes,
        display_scanned_files=verbose,
        display_binary_files=verbose,
        list_files_mode=(
            ListFilesMode.FILES_ONLY
            if not recursive
            else (
                ListFilesMode.ALL_BUT_GITIGNORED if use_gitignore else ListFilesMode.ALL
            )
        ),
    )
    target = paths[0] if len(paths) == 1 else Path.cwd()
    target_path = target if target.is_dir() else target.parent
    with ctx_obj.ui.create_scanner_ui(len(files), verbose=verbose) as scanner_ui:
        scan_context = ScanContext(
            scan_mode=ScanMode.PATH,
            command_path=ctx.command_path,
            target_path=target_path,
        )

        scanner = SecretScanner(
            client=ctx_obj.client,
            cache=ctx_obj.cache,
            ignored_matches=config.user_config.secret.ignored_matches,
            scan_context=scan_context,
            ignored_detectors=config.user_config.secret.ignored_detectors,
        )
        results = scanner.scan(files, scanner_ui=scanner_ui)
    scan = SecretScanCollection(
        id=" ".join(str(x) for x in paths), type="path_scan", results=results
    )

    return output_handler.process_scan(scan)



from abc import ABC, abstractmethod
from enum import Enum
from typing import Generic, List, Optional, TypeVar, Union

from pygitguardian.iac_models import (
    IaCDiffScanEntities,
    IaCDiffScanResult,
    IaCFileResult,
    IaCScanResult,
)


IaCResult = Union[IaCScanResult, IaCDiffScanResult]


IaCResultT = TypeVar("IaCResultT", IaCDiffScanResult, IaCScanResult)


class CollectionType(Enum):
    Unknown = "unknown"
    PathScan = "path_scan"
    DiffScan = "diff_scan"


class IaCScanCollection(ABC, Generic[IaCResultT]):
    type: CollectionType = CollectionType.Unknown
    # Can be None if the scan failed
    result: Optional[IaCResultT]

    def __init__(
        self,
        id: str,
        result: Optional[IaCResultT],
    ):
        self.id = id
        self.result = result

    @property
    @abstractmethod
    def has_results(self) -> bool:
        """
        Whether the scan found problems
        """
        raise NotImplementedError()

    @abstractmethod
    def get_entities_without_ignored(
        self,
    ) -> Optional[Union[List[IaCFileResult], IaCDiffScanEntities]]:
        """
        Removes vulnerabilities marked as ignored.
        Removes files that only have ignored vulnerabilities.
        Returns file list.
        """
        raise NotImplementedError()

    @abstractmethod
    def get_result_without_ignored(self) -> Optional[IaCResultT]:
        """
        Removes vulnerabilities marked as ignored.
        Removes files that only have ignored vulnerabilities.
        Returns result object
        """
        raise NotImplementedError()



from pathlib import Path
from typing import Any, Optional, Sequence

import click

from ggshield.cmd.sca.scan.sca_scan_utils import create_output_handler, sca_scan_all
from ggshield.cmd.sca.scan.scan_common_options import (
    add_sca_scan_common_options,
    update_context,
)
from ggshield.cmd.utils.common_decorators import exception_wrapper
from ggshield.cmd.utils.common_options import directory_argument
from ggshield.verticals.sca.collection.collection import (
    SCAScanAllVulnerabilityCollection,
)


@click.command()
@add_sca_scan_common_options()
@directory_argument
@click.pass_context
@exception_wrapper
def scan_all_cmd(
    ctx: click.Context,
    exit_zero: bool,
    minimum_severity: str,
    ignore_paths: Sequence[str],
    ignore_fixable: bool,
    ignore_not_fixable: bool,
    directory: Optional[Path],
    **kwargs: Any,
) -> int:
    """
    Scans a directory for existing vulnerabilities in open-source dependencies.

    Scanning a repository with this command will not trigger any incident on your dashboard.

    Only metadata such as call time, request size and scan mode is stored server-side.
    """
    if directory is None:
        directory = Path().resolve()

    # Adds client and required parameters to the context
    update_context(
        ctx,
        exit_zero,
        minimum_severity,
        ignore_paths,
        ignore_fixable,
        ignore_not_fixable,
    )

    result = sca_scan_all(ctx, directory)
    scan = SCAScanAllVulnerabilityCollection(id=str(directory), result=result)
    output_handler = create_output_handler(ctx)
    return output_handler.process_scan_all_result(scan)



from pathlib import Path
from typing import Any, List, Sequence

import click

from ggshield.cmd.iac.scan.all import display_iac_scan_all_result, iac_scan_all
from ggshield.cmd.iac.scan.diff import display_iac_scan_diff_result, iac_scan_diff
from ggshield.cmd.iac.scan.iac_scan_common_options import (
    add_iac_scan_common_options,
    update_context,
)
from ggshield.cmd.iac.scan.iac_scan_utils import augment_unignored_issues
from ggshield.cmd.utils.common_decorators import display_beta_warning, exception_wrapper
from ggshield.cmd.utils.common_options import all_option
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.cmd.utils.hooks import check_user_requested_skip
from ggshield.core.git_hooks.prepush import collect_commits_refs
from ggshield.core.scan.scan_mode import ScanMode
from ggshield.utils.git_shell import is_valid_git_commit_ref


@click.command()
@click.argument("prepush_args", nargs=-1, type=click.UNPROCESSED)
@add_iac_scan_common_options()
@all_option
@click.pass_context
@display_beta_warning
@exception_wrapper
def scan_pre_push_cmd(
    ctx: click.Context,
    prepush_args: List[str],
    exit_zero: bool,
    minimum_severity: str,
    ignore_policies: Sequence[str],
    ignore_paths: Sequence[str],
    scan_all: bool,
    **kwargs: Any,
) -> int:
    """
    Scan a Git repository for changes in IaC vulnerabilities in the pushed commits.
    This is intended to be used as a pre-push hook.

    The scan is successful if no *new* IaC vulnerability was found, unless `--all` is used,
    in which case the scan is only successful if no IaC vulnerability (old and new) was found.

    By default, the output will show:
    - The number of known IaC vulnerabilities resolved by the changes
    - The number of known IaC vulnerabilities left untouched
    - The number and the list of new IaC vulnerabilities introduced by the changes
    """
    if check_user_requested_skip():
        return 0

    ctx_obj = ContextObj.get(ctx)
    directory = Path().resolve()
    update_context(ctx, exit_zero, minimum_severity, ignore_policies, ignore_paths)

    _, remote_commit = collect_commits_refs(prepush_args)
    # Will happen if this is the first push on the repo
    has_no_remote_commit = remote_commit is None or not is_valid_git_commit_ref(
        remote_commit
    )

    if scan_all or has_no_remote_commit:
        result = iac_scan_all(ctx, directory, scan_mode=ScanMode.PRE_PUSH_ALL)
        augment_unignored_issues(ctx_obj.config.user_config, result)
        return display_iac_scan_all_result(ctx, directory, result)
    else:
        result = iac_scan_diff(
            ctx,
            directory,
            remote_commit,
            include_staged=False,
            scan_mode=ScanMode.PRE_PUSH_DIFF,
        )
        augment_unignored_issues(ctx_obj.config.user_config, result)
        return display_iac_scan_diff_result(ctx, directory, result)



from pathlib import Path
from typing import Any, Optional, Sequence, Union

import click
from pygitguardian.iac_models import IaCScanParameters, IaCScanResult

from ggshield.cmd.iac.scan.iac_scan_common_options import (
    add_iac_scan_common_options,
    update_context,
)
from ggshield.cmd.iac.scan.iac_scan_utils import (
    IaCSkipScanResult,
    augment_unignored_issues,
    create_output_handler,
    handle_scan_error,
)
from ggshield.cmd.utils.common_decorators import display_beta_warning, exception_wrapper
from ggshield.cmd.utils.common_options import directory_argument
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.cmd.utils.files import check_directory_not_ignored
from ggshield.core.dirs import get_project_root_dir
from ggshield.core.git_hooks.ci.supported_ci import SupportedCI
from ggshield.core.scan import ScanContext, ScanMode
from ggshield.core.text_utils import display_info
from ggshield.verticals.iac.collection.iac_path_scan_collection import (
    IaCPathScanCollection,
)
from ggshield.verticals.iac.filter import get_iac_files_from_path


# Changes to arguments must be propagated to default_command
@click.command()
@add_iac_scan_common_options()
@directory_argument
@click.pass_context
@display_beta_warning
@exception_wrapper
def scan_all_cmd(
    ctx: click.Context,
    exit_zero: bool,
    minimum_severity: str,
    ignore_policies: Sequence[str],
    ignore_paths: Sequence[str],
    directory: Optional[Path] = None,
    **kwargs: Any,
) -> int:
    """
    Scan a directory for all IaC vulnerabilities in the current state.

    The scan is successful if no IaC vulnerability (known or new) was found.
    """
    if directory is None:
        directory = Path().resolve()
    update_context(ctx, exit_zero, minimum_severity, ignore_policies, ignore_paths)

    result = iac_scan_all(ctx, directory, scan_mode=ScanMode.DIRECTORY_ALL)
    ctx_obj = ContextObj.get(ctx)
    augment_unignored_issues(ctx_obj.config.user_config, result)
    return display_iac_scan_all_result(ctx, directory, result)


def iac_scan_all(
    ctx: click.Context,
    directory: Path,
    scan_mode: ScanMode,
    ci_mode: Optional[SupportedCI] = None,
) -> Union[IaCScanResult, IaCSkipScanResult, None]:
    ctx_obj = ContextObj.get(ctx)
    config = ctx_obj.config

    check_directory_not_ignored(directory, ctx_obj.exclusion_regexes)

    paths = get_iac_files_from_path(
        path=directory,
        exclusion_regexes=ctx_obj.exclusion_regexes,
        # bypass verbose here: we want to display only IaC files
        verbose=False,
        # If the repository is a git repository, ignore untracked files
        ignore_git=False,
        ignore_git_staged=(scan_mode == ScanMode.PRE_PUSH_ALL),
    )

    if not paths:
        return IaCSkipScanResult()

    root = get_project_root_dir(directory)
    relative_paths = [str(x.resolve().relative_to(root)) for x in paths]

    if config.user_config.verbose:
        display_info("> Scanned files")
        for filepath in relative_paths:
            display_info(f"- {click.format_filename(filepath)}")

    client = ctx_obj.client

    scan_parameters = IaCScanParameters(
        list({ignored.policy for ignored in config.user_config.iac.ignored_policies}),
        config.user_config.iac.minimum_severity,
    )
    # If paths are not sorted, the tar bytes order will be different when calling the function twice
    # Different bytes order will cause different tarfile hash_key resulting in a GIM cache bypass.
    relative_paths.sort()
    scan = client.iac_directory_scan(
        root,
        relative_paths,
        scan_parameters,
        ScanContext(
            command_path=ctx.command_path,
            scan_mode=(
                scan_mode if ci_mode is None else f"{scan_mode.value}/{ci_mode.value}"
            ),
            extra_headers={"Ci-Mode": str(ci_mode.value)} if ci_mode else None,
            target_path=directory,
        ).get_http_headers(),
    )

    if not isinstance(scan, IaCScanResult):
        handle_scan_error(client, scan)
        return None
    return scan


def display_iac_scan_all_result(
    ctx: click.Context,
    directory: Path,
    result: Union[IaCScanResult, IaCSkipScanResult, None],
) -> int:
    output_handler = create_output_handler(ctx)

    if isinstance(result, IaCSkipScanResult):
        return output_handler.process_skip_scan()

    scan = IaCPathScanCollection(id=str(directory), result=result)
    return output_handler.process_scan(scan)



import json
from typing import Any, Dict, List, cast

from ggshield.verticals.sca.collection.collection import (
    SCAScanAllVulnerabilityCollection,
    SCAScanDiffVulnerabilityCollection,
)
from ggshield.verticals.sca.output.handler import SCAOutputHandler


class SCAJsonOutputHandler(SCAOutputHandler):
    def _process_scan_all_impl(self, scan: SCAScanAllVulnerabilityCollection) -> str:
        scan_dict = SCAJsonOutputHandler.create_scan_all_dict(scan)
        text = json.dumps(scan_dict)
        return cast(str, text)

    def _process_scan_diff_impl(self, scan: SCAScanDiffVulnerabilityCollection) -> str:
        result_without_ignored = scan.get_result_without_ignored()
        scan_dict = (
            result_without_ignored.to_dict()
            if result_without_ignored is not None
            else {"scanned_files": [], "added_vulns": [], "removed_vulns": []}
        )

        del scan_dict["source_found"]
        SCAJsonOutputHandler.remove_vuln_keys_from_raw_packages(
            scan_dict["added_vulns"]
        )
        SCAJsonOutputHandler.remove_vuln_keys_from_raw_packages(
            scan_dict["removed_vulns"]
        )

        return json.dumps(scan_dict)

    @staticmethod
    def create_scan_all_dict(scan: SCAScanAllVulnerabilityCollection) -> Dict[str, Any]:
        result = scan.get_result_without_ignored()
        if result is None:
            return {"scanned_files": [], "found_package_vulns": [], "total_vulns": 0}
        scan_dict = result.to_dict()

        scan_dict["total_vulns"] = 0
        for file in scan_dict["found_package_vulns"]:
            scan_dict["total_vulns"] += len(file["package_vulns"])

        del scan_dict["source_found"]
        SCAJsonOutputHandler.remove_vuln_keys_from_raw_packages(
            scan_dict["found_package_vulns"]
        )

        return scan_dict

    @staticmethod
    def remove_vuln_keys_from_raw_packages(packages: List[Dict[str, Any]]) -> None:
        """Deletes in-place elements from the raw response that should not appear in the output."""
        for package in packages:
            for package_vuln in package["package_vulns"]:
                for vuln in package_vuln["vulns"]:
                    for key_to_delete in (
                        "url",
                        "status",
                        "ignored_until",
                        "ignore_reason",
                        "ignore_comment",
                    ):
                        del vuln[key_to_delete]



from pathlib import Path
from typing import Any, Optional, Sequence

import click

from ggshield.cmd.sca.scan.sca_scan_utils import create_output_handler, sca_scan_diff
from ggshield.cmd.sca.scan.scan_common_options import (
    add_sca_scan_common_options,
    update_context,
)
from ggshield.cmd.utils.common_decorators import exception_wrapper
from ggshield.cmd.utils.common_options import (
    directory_argument,
    reference_option,
    staged_option,
)
from ggshield.core.scan.scan_mode import ScanMode
from ggshield.verticals.sca.collection.collection import (
    SCAScanDiffVulnerabilityCollection,
)


@click.command()
@add_sca_scan_common_options()
@directory_argument
@reference_option
@staged_option
@click.pass_context
@exception_wrapper
def scan_diff_cmd(
    ctx: click.Context,
    exit_zero: bool,
    minimum_severity: str,
    ignore_paths: Sequence[str],
    ignore_fixable: bool,
    ignore_not_fixable: bool,
    ref: str,
    staged: bool,
    directory: Optional[Path],
    **kwargs: Any,
) -> int:
    """
    Scans if the current revision of a git repository introduces SCA vulnerabilities.

    This command checks if the current revision introduces new vulnerabilities compared
    to the revision from GIT_REF.

    Scanning a repository with this command will not trigger any incident on your
    dashboard.

    Only metadata such as call time, request size and scan mode is stored server-side.
    """
    if directory is None:
        directory = Path().resolve()

    # Adds client and required parameters to the context
    update_context(
        ctx,
        exit_zero,
        minimum_severity,
        ignore_paths,
        ignore_fixable,
        ignore_not_fixable,
    )

    output_handler = create_output_handler(ctx)

    result = sca_scan_diff(
        ctx,
        directory,
        previous_ref=ref,
        include_staged=staged,
        scan_mode=ScanMode.DIFF,
    )
    scan = SCAScanDiffVulnerabilityCollection(id=str(directory), result=result)
    return output_handler.process_scan_diff_result(scan)



from functools import wraps
from typing import Callable, TypeVar

import click
from typing_extensions import ParamSpec

from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.core.errors import handle_exception
from ggshield.core.text_utils import display_warning


T = TypeVar("T")
P = ParamSpec("P")


def exception_wrapper(func: Callable[P, int]) -> Callable[P, int]:
    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> int:
        try:
            return func(*args, **kwargs)
        except Exception as error:
            ctx = next(arg for arg in args if isinstance(arg, click.Context))
            config = ContextObj.get(ctx).config
            return handle_exception(error, config.user_config.verbose)

    return wrapper


def display_beta_warning(func: Callable[P, T]) -> Callable[P, T]:
    """
    Displays warning about new verticals' commands being in beta.
    """

    @wraps(func)
    def func_with_beta_warning(*args: P.args, **kwargs: P.kwargs) -> T:
        display_warning(
            "This feature is still in beta, its behavior may change in future versions."
        )
        return func(*args, **kwargs)

    return func_with_beta_warning



import logging
import re
from dataclasses import field
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Tuple, Union

import marshmallow_dataclass
from marshmallow import ValidationError, post_load, pre_load

from ggshield.core.config.utils import (
    find_global_config_path,
    find_local_config_path,
    load_yaml_dict,
    remove_common_dict_items,
    replace_dash_in_keys,
    save_yaml_dict,
    update_dict_from_other,
)
from ggshield.core.config.v1_config import convert_v1_config_dict
from ggshield.core.constants import DEFAULT_LOCAL_CONFIG_PATH
from ggshield.core.errors import ParseError, UnexpectedError, format_validation_error
from ggshield.core.text_utils import display_warning
from ggshield.core.types import FilteredConfig, IgnoredMatch


logger = logging.getLogger(__name__)
CURRENT_CONFIG_VERSION = 2

_IGNORE_KNOWN_SECRETS_KEY = "ignore_known_secrets"

GHSA_ID_PATTERN = re.compile("GHSA(-[a-zA-Z0-9]{4}){3}")
CVE_ID_PATTERN = re.compile(r"CVE-\d{4}-\d{4,}")
POLICY_ID_PATTERN = re.compile("GG_IAC_[0-9]{4}")


@marshmallow_dataclass.dataclass
class SecretConfig(FilteredConfig):
    """
    Holds all user-defined secret-specific settings
    """

    show_secrets: bool = False
    ignored_detectors: Set[str] = field(default_factory=set)
    ignored_matches: List[IgnoredMatch] = field(default_factory=list)
    ignored_paths: Set[str] = field(default_factory=set)
    ignore_known_secrets: bool = False
    prereceive_remediation_message: str = """A pre-receive hook set server side prevented you from pushing secrets.
Since the secret was detected during the push BUT after the commit, you need to:
1. rewrite the git history making sure to replace the secret with its reference (e.g. environment variable).
2. push again."""

    def add_ignored_match(self, secret: IgnoredMatch) -> None:
        """
        Add secret to ignored_matches.
        """
        for match in self.ignored_matches:
            if match.match == secret.match:
                # take the opportunity to name the ignored match
                if not match.name:
                    match.name = secret.name
                return
        self.ignored_matches.append(secret)


def validate_policy_id(policy_id: str) -> bool:
    return bool(POLICY_ID_PATTERN.fullmatch(policy_id))


@marshmallow_dataclass.dataclass
class ConfigIgnoredElement(FilteredConfig):
    """
    Base class for a config element which can be ignored temporarily.
    Classes inheriting it must reject expired dates themselves.
    """

    comment: Optional[str]
    until: Optional[datetime]

    # Accept date yyyy-mm-dd instead of a full datetime
    @pre_load
    def parse_date(self, data: Dict[str, Any], **kwargs: Any):
        if not (isinstance(data, dict)) or data.get("until") is None:
            return data
        try:
            data["until"] = str(datetime.strptime(str(data["until"]), "%Y-%m-%d"))
            return data
        except ValueError:
            return data

    @post_load
    def datetime_to_utc(self, data: Dict[str, Any], **kwargs: Any):
        if data["until"] is not None:
            data["until"] = data["until"].astimezone(timezone.utc)
        return data


def remove_expired_elements(
    lst: List[ConfigIgnoredElement],
) -> List[ConfigIgnoredElement]:
    expired: List[ConfigIgnoredElement] = []
    now = datetime.now(tz=timezone.utc)
    for idx in range(len(lst) - 1, -1, -1):
        ignored = lst[idx]
        if ignored.until and ignored.until <= now:
            lst.pop(idx)
            expired.insert(0, ignored)

    return expired


def report_expired_elements(expired_lst: List[ConfigIgnoredElement]) -> None:
    for element in expired_lst:
        display_warning(
            f"{element} has an expired 'until' "
            f"date ({element.until}), please update your configuration file."
        )


@marshmallow_dataclass.dataclass
class IaCConfigIgnoredPath(ConfigIgnoredElement):
    path: str

    def __init__(
        self,
        path: str,
        comment: Optional[str] = None,
        until: Optional[datetime] = None,
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(comment, until, *args, **kwargs)
        self.path = path

    def __str__(self):
        return f"Path {self.path}"

    @pre_load
    def convert_paths(self, in_data: Union[str, Dict[str, Any]], **kwargs: Any):
        return {"path": in_data} if isinstance(in_data, str) else in_data


@marshmallow_dataclass.dataclass
class IaCConfigIgnoredPolicy(ConfigIgnoredElement):
    policy: str = field(metadata={"validate": validate_policy_id})

    def __init__(
        self,
        policy: str,
        comment: Optional[str] = None,
        until: Optional[datetime] = None,
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(comment, until, *args, **kwargs)
        self.policy = policy

    def __str__(self):
        return f"Policy {self.policy}"

    @pre_load
    def convert_policies(self, in_data: Union[str, Dict[str, Any]], **kwargs: Any):
        return {"policy": in_data} if isinstance(in_data, str) else in_data


@marshmallow_dataclass.dataclass
class IaCConfig(FilteredConfig):
    """
    Holds the iac config as defined .gitguardian.yaml files
    (local and global).
    """

    ignored_paths: List[IaCConfigIgnoredPath] = field(default_factory=list)
    ignored_policies: List[IaCConfigIgnoredPolicy] = field(default_factory=list)
    minimum_severity: str = "LOW"

    # If we hit any outdated ignore rules when loading the configuration file, we
    # want to keep them in order to display additional details but we do not want
    # to store them in ignored rules because they will be used to scan.
    # Instead, we store them in these new computed fields. However we must mark them
    # as excluded, otherwise they would be serialized if the user runs a command to
    # write in the configuration file, which would be odd.
    outdated_ignored_paths: List[IaCConfigIgnoredPath] = field(default_factory=list)
    outdated_ignored_policies: List[IaCConfigIgnoredPolicy] = field(
        default_factory=list
    )

    @post_load
    def validate_ignored_paths(self, data: Dict[str, Any], **kwargs: Any):
        expired_lst = remove_expired_elements(data["ignored_paths"])
        report_expired_elements(expired_lst)
        data["outdated_ignored_paths"] = expired_lst
        return data

    @post_load
    def validate_ignored_policies(self, data: Dict[str, Any], **kwargs: Any):
        expired_lst = remove_expired_elements(data["ignored_policies"])
        report_expired_elements(expired_lst)
        data["outdated_ignored_policies"] = expired_lst
        return data


def is_ghsa_valid(ghsa_id: str) -> bool:
    return bool(GHSA_ID_PATTERN.fullmatch(ghsa_id))


def is_cve_id_valid(cve_id: str) -> bool:
    return bool(CVE_ID_PATTERN.fullmatch(cve_id))


def validate_vuln_identifier(value: str):
    if not (is_ghsa_valid(value) or is_cve_id_valid(value)):
        raise ValidationError(
            f"The given identifier '{value}' do not match any of the allowed patterns"
            f" '{GHSA_ID_PATTERN.pattern}' or '{CVE_ID_PATTERN.pattern}'"
        )


@marshmallow_dataclass.dataclass
class SCAConfigIgnoredVulnerability(ConfigIgnoredElement):
    """
    A model of an ignored vulnerability for SCA. This allows to ignore all occurrences
    of a given vulnerability in a given dependency file.
    - identifier: identifier (currently: either CVE id or GHSA id) of the vulnerability to ignore
    - path: the path to the file in which ignore the vulnerability
    - comment: The ignored reason
    - until: A datetime until which the vulnerability is ignored
    """

    identifier: str = field(metadata={"validate": validate_vuln_identifier})
    path: str

    def __init__(
        self,
        identifier: str,
        path: str,
        comment: Optional[str] = None,
        until: Optional[datetime] = None,
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(comment, until, *args, **kwargs)
        self.identifier = identifier
        self.path = path

    def __str__(self):
        return f"Vulnerability {self.identifier}"


@marshmallow_dataclass.dataclass
class SCAConfig(FilteredConfig):
    """
    Holds the sca config as defined .gitguardian.yaml files
    (local and global).
    """

    ignored_paths: Set[str] = field(default_factory=set)
    minimum_severity: str = "LOW"
    ignored_vulnerabilities: List[SCAConfigIgnoredVulnerability] = field(
        default_factory=list
    )
    ignore_not_fixable: bool = False
    ignore_fixable: bool = False

    @post_load
    def validate_ignored_vulns(self, data: Dict[str, Any], **kwargs: Any):
        expired_lst = remove_expired_elements(data["ignored_vulnerabilities"])
        report_expired_elements(expired_lst)
        return data


@marshmallow_dataclass.dataclass
class UserConfig(FilteredConfig):
    """
    Holds all ggshield settings defined by the user in the .gitguardian.yaml files
    (local and global).
    """

    iac: IaCConfig = field(default_factory=IaCConfig)
    instance: Optional[str] = None
    exit_zero: bool = False
    verbose: bool = False
    allow_self_signed: bool = False
    max_commits_for_hook: int = 50
    sca: SCAConfig = field(default_factory=SCAConfig)
    secret: SecretConfig = field(default_factory=SecretConfig)
    debug: bool = False

    # If we hit any deprecated syntax when loading a configuration file, we do not
    # display them directly, otherwise the messages would also be shown when running
    # `ggshield config migrate`, which would be odd.
    # Instead, we store them in this list and a result_callback() function displays
    # them when we quit. When `config migrate` runs, it clears this list, so nothing
    # gets displayed.
    deprecation_messages: List[str] = field(default_factory=list)

    def save(self, config_path: Path) -> None:
        """
        Save config to config_path
        """
        dct = self.to_config_dict()
        save_yaml_dict(dct, config_path)

    def to_config_dict(self) -> Dict[str, Any]:
        dct = self.to_dict()
        default_dct = UserConfig.from_dict({}).to_dict()

        dct = remove_common_dict_items(dct, default_dct)

        dct["version"] = CURRENT_CONFIG_VERSION

        return dct

    @classmethod
    def load(cls, config_path: Optional[Path] = None) -> Tuple["UserConfig", Path]:
        """
        Load the various user configs files to create a UserConfig object:
        - global user configuration file (in the home)
        - local user configuration file (in the repository)

        Returns a UserConfig instance, and the path where updates should be saved
        """
        deprecation_messages: List[str] = []
        if config_path:
            logger.debug("Loading custom config from %s", config_path)
            dct = _load_config_dict(config_path, deprecation_messages)
            user_config = UserConfig.from_config_dict(dct)
            user_config.deprecation_messages = deprecation_messages
            return user_config, config_path

        user_config_dict: Dict[str, Any] = {}
        global_config_path = find_global_config_path()
        if global_config_path:
            dct = _load_config_dict(global_config_path, deprecation_messages)
            update_dict_from_other(user_config_dict, dct)
            logger.debug("Loaded global config from %s", global_config_path)
        else:
            logger.debug("No global config")

        local_config_path = find_local_config_path()
        if local_config_path:
            dct = _load_config_dict(local_config_path, deprecation_messages)
            update_dict_from_other(user_config_dict, dct)
            config_path = local_config_path
            logger.debug("Loaded local config from %s", local_config_path)
        else:
            logger.debug("No local config")

        user_config = UserConfig.from_config_dict(user_config_dict)
        user_config.deprecation_messages = deprecation_messages

        if config_path is None:
            config_path = Path(DEFAULT_LOCAL_CONFIG_PATH)
        return user_config, config_path

    @staticmethod
    def from_config_dict(data: Dict[str, Any]) -> "UserConfig":
        """Create a UserConfig instance. In case of error, format it and raise
        ParseError."""
        try:
            return UserConfig.from_dict(data)
        except ValidationError as exc:
            message = format_validation_error(exc)
            raise ParseError(message) from exc


UserConfig.SCHEMA = marshmallow_dataclass.class_schema(UserConfig)(
    exclude=(
        "deprecation_messages",
        "iac.outdated_ignored_paths",
        "iac.outdated_ignored_policies",
    )
)


def _load_config_dict(
    config_path: Path, deprecation_messages: List[str]
) -> Dict[str, Any]:
    """
    Load configuration from `config_path` as a dict.
    Appends any deprecation message regarding this file to `deprecation_messages`.
    """
    try:
        # load_yaml_dict() returns None if `config_path` does not exist. When this
        # happens, initialize the config dict to an empty configuration file,
        # but set the version to the latest version so that we don't get deprecation
        # messages.
        dct = load_yaml_dict(config_path) or {"version": CURRENT_CONFIG_VERSION}
    except ValueError as exc:
        raise ParseError(str(exc)) from exc

    dash_keys = replace_dash_in_keys(dct)

    config_version = dct.pop("version", 1)
    if config_version == 2:
        if dash_keys:
            _warn_about_dash_keys(config_path, dash_keys)
        _fix_ignore_known_secrets(dct)
    elif config_version == 1:
        deprecation_messages.append(
            f"{config_path} uses a deprecated configuration file format."
            " Run `ggshield config migrate` to migrate it to the latest version."
        )
        dct = convert_v1_config_dict(dct, deprecation_messages)
    else:
        raise UnexpectedError(f"Don't know how to load config version {config_version}")
    return dct


def _fix_ignore_known_secrets(data: Dict[str, Any]) -> None:
    """Fix a mistake done when implementing ignore-known-secrets: since this is a secret
    specific key, it should have been stored in the "secret" mapping, not in the root
    one"""

    value = data.pop(_IGNORE_KNOWN_SECRETS_KEY, None)
    if value is None:
        # No key to fix
        return

    secret_dct = data.setdefault("secret", {})
    if _IGNORE_KNOWN_SECRETS_KEY in secret_dct:
        # Do not override if it's already there
        return
    secret_dct[_IGNORE_KNOWN_SECRETS_KEY] = value


def _warn_about_dash_keys(config_path: Path, dash_keys: Set[str]) -> None:
    for old_key in sorted(dash_keys):
        new_key = old_key.replace("-", "_")
        display_warning(
            f"{config_path}: Config key {old_key} is deprecated, use {new_key} instead."
        )



from typing import List

from pygitguardian.sca_models import (
    SCALocationVulnerability,
    SCAVulnerability,
    SCAVulnerablePackageVersion,
)


def filter_unignored_vulnerabilities(
    vulnerabilities: List[SCAVulnerability],
) -> List[SCAVulnerability]:
    """Removes ignored vulnerabilities from the given list"""
    return [vuln for vuln in vulnerabilities if vuln.status != "IGNORED"]


def filter_unignored_location_vulnerabilities(
    locations: List[SCALocationVulnerability],
) -> List[SCALocationVulnerability]:
    """
    Removes all ignored vulnerabilities from all locations.
    Removes locations and nested objects if they are empty.
    """
    unignored_locations: List[SCALocationVulnerability] = []
    for location in locations:
        unignored_package_vulns: List[SCAVulnerablePackageVersion] = []
        for package_vuln in location.package_vulns:
            unignored_vulns = filter_unignored_vulnerabilities(package_vuln.vulns)
            if len(unignored_vulns) > 0:
                unignored_package_vulns.append(
                    SCAVulnerablePackageVersion(
                        package_full_name=package_vuln.package_full_name,
                        version=package_vuln.version,
                        ecosystem=package_vuln.ecosystem,
                        dependency_type=package_vuln.dependency_type,
                        vulns=unignored_vulns,
                    )
                )
        if len(unignored_package_vulns) > 0:
            unignored_locations.append(
                SCALocationVulnerability(
                    location=location.location,
                    package_vulns=unignored_package_vulns,
                )
            )
    return unignored_locations



from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, Iterable, List, NamedTuple, Optional, Tuple, Union, cast

from pygitguardian.models import Match, ScanResult

from ggshield.core.errors import UnexpectedError
from ggshield.core.filter import group_policy_breaks_by_ignore_sha
from ggshield.core.lines import Line, get_lines_from_content
from ggshield.core.scan.scannable import Scannable
from ggshield.utils.git_shell import Filemode
from ggshield.verticals.secret.extended_match import ExtendedMatch


class Result:
    """
    Return model for a scan which zips the information
    between the Scan result and its input file.
    """

    filename: str  # Name of the file/patch scanned
    filemode: Filemode
    path: Path
    url: str
    scan: ScanResult  # Result of content scan

    def __init__(self, file: Scannable, scan: ScanResult):
        self.filename = file.filename
        self.filemode = file.filemode
        self.path = file.path
        self.url = file.url
        self.scan = scan
        lines = get_lines_from_content(file.content, self.filemode)
        self.enrich_matches(lines)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Result):
            return False
        return (
            self.filename == other.filename
            and self.filemode == other.filemode
            and self.path == other.path
            and self.url == other.url
            and self.scan == other.scan
        )

    @property
    def is_on_patch(self) -> bool:
        return self.filemode != Filemode.FILE

    def enrich_matches(self, lines: List[Line]) -> None:
        if len(lines) == 0:
            raise UnexpectedError("Parsing of scan result failed.")
        for policy_break in self.scan.policy_breaks:
            policy_break.matches = cast(
                List[Match],
                [
                    ExtendedMatch.from_match(match, lines, self.is_on_patch)
                    for match in policy_break.matches
                ],
            )

    def censor(self) -> None:
        for policy_break in self.scan.policy_breaks:
            for extended_match in policy_break.matches:
                cast(ExtendedMatch, extended_match).censor()

    @property
    def has_policy_breaks(self) -> bool:
        return self.scan.has_policy_breaks


class Error(NamedTuple):
    files: List[Tuple[str, Filemode]]
    description: str  # Description of the error


@dataclass
class Results:
    """
    Return model for a scan with the results and errors of the scan

    Not a NamedTuple like the others because it causes mypy 0.961 to crash on the
    `from_exception()` method (!)

    Similar crash: https://github.com/python/mypy/issues/12629
    """

    results: List[Result] = field(default_factory=list)
    errors: List[Error] = field(default_factory=list)

    @staticmethod
    def from_exception(exc: Exception) -> "Results":
        """Create a Results representing a failure"""
        error = Error(files=[], description=str(exc))
        return Results(results=[], errors=[error])

    def extend(self, others: "Results") -> None:
        self.results.extend(others.results)
        self.errors.extend(others.errors)

    @property
    def has_policy_breaks(self) -> bool:
        return any(x.has_policy_breaks for x in self.results)


class SecretScanCollection:
    id: str
    type: str
    results: Optional[Results] = None
    scans: Optional[List["SecretScanCollection"]] = None
    optional_header: Optional[str] = None  # To be printed in Text Output
    extra_info: Optional[Dict[str, str]] = None  # To be included in JSON Output

    def __init__(
        self,
        id: Union[str, Path],
        type: str,
        results: Optional[Results] = None,
        scans: Optional[List["SecretScanCollection"]] = None,
        optional_header: Optional[str] = None,
        extra_info: Optional[Dict[str, str]] = None,
    ):
        self.id = str(id)
        self.type = type
        self.results = results
        self.scans = scans
        self.optional_header = optional_header
        self.extra_info = extra_info

        (
            self.known_secrets_count,
            self.new_secrets_count,
        ) = self._get_known_new_secrets_count()

    @property
    def has_new_secrets(self) -> bool:
        return self.new_secrets_count > 0

    @property
    def has_secrets(self) -> bool:
        return (self.new_secrets_count + self.known_secrets_count) > 0

    @property
    def scans_with_results(self) -> List["SecretScanCollection"]:
        if self.scans:
            return [scan for scan in self.scans if scan.results]
        return []

    @property
    def has_results(self) -> bool:
        return bool(self.results and self.results.results)

    def _get_known_new_secrets_count(self) -> Tuple[int, int]:
        policy_breaks = []
        for result in self.get_all_results():
            for policy_break in result.scan.policy_breaks:
                policy_breaks.append(policy_break)

        known_secrets_count = 0
        new_secrets_count = 0
        sha_dict = group_policy_breaks_by_ignore_sha(policy_breaks)

        for ignore_sha, policy_breaks in sha_dict.items():
            if policy_breaks[0].known_secret:
                known_secrets_count += 1
            else:
                new_secrets_count += 1

        return known_secrets_count, new_secrets_count

    def get_all_results(self) -> Iterable[Result]:
        """Returns an iterable on all results and sub-scan results"""
        if self.results:
            yield from self.results.results
        if self.scans:
            for scan in self.scans:
                if scan.results:
                    yield from scan.results.results



from marshmallow import fields
from pygitguardian.iac_models import (
    IaCDiffScanResultSchema,
    IaCFileResultSchema,
    IaCScanResultSchema,
    IaCVulnerabilitySchema,
)
from pygitguardian.models import BaseSchema


class IaCJSONVulnerabilitySchema(IaCVulnerabilitySchema):
    class Meta(IaCVulnerabilitySchema.Meta):
        exclude = ("url", "status", "ignored_until", "ignore_reason", "ignore_comment")


class IaCJSONFileResultSchema(IaCFileResultSchema):
    incidents = fields.List(fields.Nested(IaCJSONVulnerabilitySchema))
    total_incidents = fields.Integer(dump_default=0)


class IaCJSONScanResultSchema(IaCScanResultSchema):
    entities_with_incidents = fields.List(fields.Nested(IaCJSONFileResultSchema))
    total_incidents = fields.Integer(dump_default=0)

    class Meta(IaCScanResultSchema.Meta):
        exclude = ("source_found",)


class IaCJSONScanDiffEntitiesSchema(BaseSchema):
    unchanged = fields.List(fields.Nested(IaCJSONFileResultSchema))
    new = fields.List(fields.Nested(IaCJSONFileResultSchema))
    deleted = fields.List(fields.Nested(IaCJSONFileResultSchema))


class IaCJSONScanDiffResultSchema(IaCDiffScanResultSchema):
    entities_with_incidents = fields.Nested(IaCJSONScanDiffEntitiesSchema)

    class Meta(IaCDiffScanResultSchema.Meta):
        exclude = ("source_found",)



from pathlib import Path
from typing import Any, List

import click

from ggshield.cmd.secret.scan.secret_scan_common_options import (
    add_secret_scan_common_options,
)
from ggshield.cmd.utils.common_decorators import exception_wrapper
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.cmd.utils.hooks import check_user_requested_skip
from ggshield.core.scan import Commit, ScanContext, ScanMode
from ggshield.utils.git_shell import check_git_dir
from ggshield.verticals.secret import SecretScanCollection, SecretScanner
from ggshield.verticals.secret.output import SecretTextOutputHandler
from ggshield.verticals.secret.output.messages import remediation_message


REMEDIATION_STEPS = """Since the secret was detected before the commit was made:
1. replace the secret with its reference (e.g. environment variable).
2. commit again."""

BYPASS_MESSAGE = """  - if you use the pre-commit framework:

     SKIP=ggshield git commit -m "<your message>"

  - otherwise (warning: the following command bypasses all pre-commit hooks):

     git commit -m "<your message>" --no-verify"""


@click.command()
@click.argument("precommit_args", nargs=-1, type=click.UNPROCESSED)
@add_secret_scan_common_options()
@click.pass_context
@exception_wrapper
def precommit_cmd(
    ctx: click.Context, precommit_args: List[str], **kwargs: Any
) -> int:  # pragma: no cover
    """
    Scan as a pre-commit hook all changes that have been staged in a git repository.
    """
    ctx_obj = ContextObj.get(ctx)
    config = ctx_obj.config
    verbose = config.user_config.verbose

    if check_user_requested_skip():
        return 0

    output_handler = SecretTextOutputHandler(
        show_secrets=config.user_config.secret.show_secrets,
        verbose=verbose,
        output=None,
        ignore_known_secrets=config.user_config.secret.ignore_known_secrets,
    )
    check_git_dir()

    scan_context = ScanContext(
        scan_mode=ScanMode.PRE_COMMIT,
        command_path=ctx.command_path,
        target_path=Path.cwd(),
    )

    commit = Commit.from_staged(ctx_obj.exclusion_regexes)
    scanner = SecretScanner(
        client=ctx_obj.client,
        cache=ctx_obj.cache,
        scan_context=scan_context,
        ignored_matches=config.user_config.secret.ignored_matches,
        ignored_detectors=config.user_config.secret.ignored_detectors,
    )
    with ctx_obj.ui.create_scanner_ui(len(commit.urls), verbose=verbose) as scanner_ui:
        results = scanner.scan(commit.get_files(), scanner_ui)

    return_code = output_handler.process_scan(
        SecretScanCollection(id="cached", type="pre-commit", results=results)
    )
    if return_code:
        click.echo(
            remediation_message(
                remediation_steps=REMEDIATION_STEPS,
                bypass_message=BYPASS_MESSAGE,
                rewrite_git_history=False,
            ),
            err=True,
        )
    return return_code



from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, Optional, Pattern, Set, Type, Union

import click
from pygitguardian import GGClient
from pygitguardian.iac_models import IaCDiffScanResult, IaCScanResult
from pygitguardian.models import Detail

from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.core.config.user_config import UserConfig
from ggshield.core.errors import APIKeyCheckError
from ggshield.core.filter import init_exclusion_regexes
from ggshield.core.tar_utils import INDEX_REF, tar_from_ref_and_filepaths
from ggshield.core.text_utils import display_error
from ggshield.utils.files import is_path_excluded
from ggshield.utils.git_shell import get_filepaths_from_ref, get_staged_filepaths
from ggshield.verticals.iac.collection.iac_scan_collection import IaCResult
from ggshield.verticals.iac.filter import is_iac_file_path
from ggshield.verticals.iac.output import (
    IaCJSONOutputHandler,
    IaCOutputHandler,
    IaCTextOutputHandler,
)


@dataclass
class IaCSkipScanResult:
    id: str = ""


def create_output_handler(ctx: click.Context) -> IaCOutputHandler:
    """Read objects defined in ctx.obj and create the appropriate OutputHandler
    instance"""
    ctx_obj = ContextObj.get(ctx)
    output_handler_cls: Type[IaCOutputHandler]
    if ctx_obj.use_json:
        output_handler_cls = IaCJSONOutputHandler
    else:
        output_handler_cls = IaCTextOutputHandler
    return output_handler_cls(verbose=ctx_obj.config.user_config.verbose)


def handle_scan_error(client: GGClient, detail: Detail) -> None:
    if detail.status_code == 401:
        raise APIKeyCheckError(client.base_uri, "Invalid API key.")
    display_error("\nError scanning.")
    display_error(str(detail))


def get_git_filepaths(directory: Path, ref: str) -> Iterable[Path]:
    return (
        get_staged_filepaths(str(directory))
        if ref == INDEX_REF
        else get_filepaths_from_ref(ref, str(directory))
    )


def _accept_iac_file_on_path(
    path: Path, directory: Path, exclusion_regexes: Optional[Set[Pattern[str]]] = None
) -> bool:
    return is_iac_file_path(path) and (
        exclusion_regexes is None
        or not is_path_excluded(directory / path, exclusion_regexes)
    )


def filter_iac_filepaths(
    directory: Path,
    filepaths: Iterable[Path],
    exclusion_regexes: Optional[Set[Pattern[str]]] = None,
) -> Iterable[Path]:
    # You can filter based on file's content here
    # using read_git_file (result will be cached)
    # You should filter on path first, in order to read
    # the content only if necessary
    return [
        path
        for path in filepaths
        if _accept_iac_file_on_path(
            path, directory, exclusion_regexes=exclusion_regexes
        )
    ]


def get_iac_tar(
    directory: Path, ref: str, exclusion_regexes: Set[Pattern[str]]
) -> bytes:
    filepaths = get_git_filepaths(directory, ref)
    filtered_paths = filter_iac_filepaths(
        directory=directory,
        filepaths=filepaths,
        exclusion_regexes=exclusion_regexes,
    )

    return tar_from_ref_and_filepaths(ref, filtered_paths, wd=str(directory))


def augment_unignored_issues(
    user_config: UserConfig,
    result: Union[IaCResult, IaCSkipScanResult, None],
) -> None:
    """
    GIVEN a list of vulnerabilities from a scan result
    WHEN ignored policies and paths are configured but outdated
    THEN augment the vulnerability with the date it was last ignored
    """
    if isinstance(result, IaCScanResult):
        incidents_list = result.entities_with_incidents
    elif isinstance(result, IaCDiffScanResult):
        incidents_list = [
            *result.entities_with_incidents.new,
            *result.entities_with_incidents.unchanged,
            *result.entities_with_incidents.deleted,
        ]
    else:
        return
    if len(incidents_list) == 0:
        return
    outdated_ignored_paths = user_config.iac.outdated_ignored_paths
    outdated_ignored_policies = user_config.iac.outdated_ignored_policies
    # Early return if there are no outdated configurations
    if len(outdated_ignored_paths) == 0 and len(outdated_ignored_policies) == 0:
        return
    outdated_ignored_paths_dicts = [
        {
            "regex": init_exclusion_regexes({outdated_ignored_path.path}),
            "until": outdated_ignored_path.until,
        }
        for outdated_ignored_path in outdated_ignored_paths
    ]
    outdated_ignored_policies_dict = {
        outdated_policy.policy: outdated_policy.until
        for outdated_policy in outdated_ignored_policies
    }
    # For each file and each vulnerability within
    for file_result in incidents_list:
        # Check if path was ignored
        file_path = file_result.filename
        file_ignored_until = None
        for outdated_ignored_path in outdated_ignored_paths_dicts:
            if is_path_excluded(file_path, outdated_ignored_path["regex"]):
                if (
                    file_ignored_until is None
                    or file_ignored_until < outdated_ignored_path["until"]
                ):
                    file_ignored_until = outdated_ignored_path["until"]
        for vulnerability in file_result.incidents:
            # Check if policy was ignored
            policy_ignored_until = outdated_ignored_policies_dict.get(
                vulnerability.policy_id
            )
            until_dates = (policy_ignored_until, file_ignored_until)
            # Augment vulnerability with the most recent ignored_until date
            vulnerability.ignored_until = max(
                (d for d in until_dates if d is not None), default=None
            )



import os
from pathlib import Path
from typing import Dict, Optional, Tuple, Union

import click

from ggshield.core.errors import NotAMergeRequestError, UnexpectedError
from ggshield.utils.git_shell import get_commits_not_in_branch, get_remotes

from .supported_ci import SupportedCI


def travis_scan_ci_args() -> Tuple[str, str]:
    if "TRAVIS_PULL_REQUEST" not in os.environ:
        raise NotAMergeRequestError()
    try:
        commit_range = os.environ["TRAVIS_COMMIT_RANGE"]
        first_commit, last_commit = commit_range.split("..")
        return last_commit, first_commit + "~1"
    except Exception as exc:
        raise UnexpectedError("Failed to extract scan ci arguments for travis") from exc


# Note: this does not exist (yet ?) for CircleCI, see
# https://circleci.canny.io/config/p/provide-env-variable-for-branch-name-targeted-by-pull-request
CI_TARGET_BRANCH_ASSOC: Dict[SupportedCI, str] = {
    SupportedCI.GITHUB: "GITHUB_BASE_REF",
    SupportedCI.GITLAB: "CI_MERGE_REQUEST_TARGET_BRANCH_NAME",
    SupportedCI.JENKINS: "CHANGE_TARGET",
    SupportedCI.AZURE: "SYSTEM_PULLREQUEST_TARGETBRANCHNAME",
    SupportedCI.BITBUCKET: "BITBUCKET_PR_DESTINATION_BRANCH",
    SupportedCI.DRONE: "DRONE_COMMIT_BRANCH",
}


def get_scan_ci_parameters(
    current_ci: SupportedCI,
    wd: Optional[Union[str, Path]] = None,
    verbose: bool = False,
) -> Union[Tuple[str, str], None]:
    """
    Function used to gather current commit and reference commit, for the SCA/IaC scan
    ci commands.
    Return:
      - a tuple (current_commit, reference commit) in the nominal case
      - or None if the MR has no associated commits
        (i.e. the mr branch has no new commit compared to the target branch)

    Note: this function will not work (i.e. probably raise) if the git directory is a shallow clone
    """
    if verbose:
        click.echo(
            f"\tIdentified current ci as {current_ci.value}",
            err=True,
        )

    if current_ci == SupportedCI.TRAVIS:
        return travis_scan_ci_args()

    remotes = get_remotes(wd=wd)
    if len(remotes) == 0:
        # note: this should not happen in practice, esp. in a CI job
        if verbose:
            click.echo(
                "\tNo remote found.",
                err=True,
            )
        remote_prefix = ""
    else:
        if verbose:
            click.echo(
                f"\tUsing first remote {remotes[0]}.",
                err=True,
            )
        remote_prefix = f"{remotes[0]}/"

    target_branch_var = CI_TARGET_BRANCH_ASSOC.get(current_ci)
    if not target_branch_var:
        raise UnexpectedError(f"Using scan ci is not supported for {current_ci.value}.")

    if not os.getenv(target_branch_var):
        raise NotAMergeRequestError()
    target_branch = remote_prefix + os.environ[target_branch_var]

    mr_tip = "HEAD"
    if current_ci == SupportedCI.GITHUB:
        # - GitHub pipelines are all merge results
        # - GITHUB_BASE_REF and GITHUB_HEAD_REF are always set together
        #   "when the event that triggers a workflow run is either pull_request or pull_request_target"
        mr_tip = remote_prefix + os.environ["GITHUB_HEAD_REF"]
    if current_ci == SupportedCI.GITLAB:
        # Handles gitlab's MR, whether merge results pipelines are enabled or not
        mr_tip = remote_prefix + os.environ["CI_MERGE_REQUEST_SOURCE_BRANCH_NAME"]

    mr_commits = get_commits_not_in_branch(
        current_tip=mr_tip, target_branch=target_branch, wd=wd
    )
    if len(mr_commits) == 0:
        return None

    current_commit = mr_commits[0]
    reference_commit = mr_commits[-1] + "~1"
    if verbose:
        click.echo(
            (
                f"\tIdentified current commit as {current_commit}\n"
                f"\tIdentified reference commit as {reference_commit}"
            ),
            err=True,
        )

    return current_commit, reference_commit



from pathlib import Path
from typing import Any

import click

from ggshield.cmd.secret.scan.secret_scan_common_options import (
    add_secret_scan_common_options,
    create_output_handler,
)
from ggshield.cmd.utils.common_decorators import exception_wrapper
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.core.scan import ScanContext, ScanMode
from ggshield.core.text_utils import display_info, pluralize
from ggshield.utils.git_shell import (
    check_git_dir,
    get_default_branch,
    get_list_commit_SHA,
)
from ggshield.verticals.secret.repo import scan_commit_range


@click.command()
@add_secret_scan_common_options()
@click.pass_context
@exception_wrapper
def changes_cmd(ctx: click.Context, **kwargs: Any) -> int:
    """
    Scan the set of changes between the default branch and the current HEAD, including staged changes.
    """
    ctx_obj = ContextObj.get(ctx)
    config = ctx_obj.config
    check_git_dir()

    default_branch = get_default_branch()
    commit_list = get_list_commit_SHA(f"{default_branch}..HEAD")

    if config.user_config.verbose:
        display_info(
            f"Scan staged changes and {len(commit_list)} new {pluralize('commit', len(commit_list))}"
        )

    scan_context = ScanContext(
        scan_mode=ScanMode.CHANGE,
        command_path=ctx.command_path,
        target_path=Path.cwd(),
    )

    return scan_commit_range(
        client=ctx_obj.client,
        cache=ctx_obj.cache,
        ui=ctx_obj.ui,
        commit_list=commit_list,
        output_handler=create_output_handler(ctx),
        exclusion_regexes=ctx_obj.exclusion_regexes,
        matches_ignore=config.user_config.secret.ignored_matches,
        scan_context=scan_context,
        ignored_detectors=config.user_config.secret.ignored_detectors,
        verbose=config.user_config.verbose,
        include_staged=True,
    )
